"""Supervision v0.6.0 with Optional xyxy enabled"""

from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Iterator, List, Optional, Tuple

import numpy as np

from supervision.detection.utils import non_max_suppression, xywh_to_xyxy
from supervision.geometry.core import Position


def calculate_masks_centroids(masks: np.ndarray) -> np.ndarray:
    """
    Calculate the centroids of binary masks in a tensor.

    Parameters:
        masks (np.ndarray): A 3D NumPy array of shape (num_masks, height, width).
            Each 2D array in the tensor represents a binary mask.

    Returns:
        A 2D NumPy array of shape (num_masks, 2), where each row contains the x and y
            coordinates (in that order) of the centroid of the corresponding mask.
    """
    num_masks, height, width = masks.shape
    total_pixels = masks.sum(axis=(1, 2))

    # offset for 1-based indexing
    vertical_indices, horizontal_indices = np.indices((height, width)) + 0.5
    # avoid division by zero for empty masks
    total_pixels[total_pixels == 0] = 1

    def sum_over_mask(indices: np.ndarray, axis: tuple) -> np.ndarray:
        return np.tensordot(masks, indices, axes=axis)

    aggregation_axis = ([1, 2], [0, 1])
    centroid_x = sum_over_mask(horizontal_indices, aggregation_axis) / total_pixels
    centroid_y = sum_over_mask(vertical_indices, aggregation_axis) / total_pixels

    return np.column_stack((centroid_x, centroid_y)).astype(int)


def _validate_xyxy(xyxy: Any, n: int) -> None:
    is_valid = isinstance(xyxy, np.ndarray) and xyxy.shape == (n, 4)
    if not is_valid:
        raise ValueError("xyxy must be 2d np.ndarray with (n, 4) shape")


def _validate_mask(mask: Any, n: int) -> None:
    is_valid = mask is None or (
        isinstance(mask, np.ndarray) and len(mask.shape) == 3 and mask.shape[0] == n
    )
    if not is_valid:
        raise ValueError("mask must be 3d np.ndarray with (n, W, H) shape")


def _validate_class_id(class_id: Any, n: int) -> None:
    is_valid = class_id is None or (
        isinstance(class_id, np.ndarray) and class_id.shape == (n,)
    )
    if not is_valid:
        raise ValueError("class_id must be None or 1d np.ndarray with (n,) shape")


def _validate_confidence(confidence: Any, n: int) -> None:
    is_valid = confidence is None or (
        isinstance(confidence, np.ndarray) and confidence.shape == (n,)
    )
    if not is_valid:
        raise ValueError("confidence must be None or 1d np.ndarray with (n,) shape")


def _validate_tracker_id(tracker_id: Any, n: int) -> None:
    is_valid = tracker_id is None or (
        isinstance(tracker_id, np.ndarray) and tracker_id.shape == (n,)
    )
    if not is_valid:
        raise ValueError("tracker_id must be None or 1d np.ndarray with (n,) shape")


@dataclass
class Detections:
    """
    Data class containing information about the detections in a video frame.
    Attributes:
        xyxy (np.ndarray): An array of shape `(n, 4)` containing the bounding boxes coordinates in format `[x1, y1, x2, y2]`
        mask: (Optional[np.ndarray]): An array of shape `(n, W, H)` containing the segmentation masks.
        confidence (Optional[np.ndarray]): An array of shape `(n,)` containing the confidence scores of the detections.
        class_id (Optional[np.ndarray]): An array of shape `(n,)` containing the class ids of the detections.
        tracker_id (Optional[np.ndarray]): An array of shape `(n,)` containing the tracker ids of the detections.
    """

    xyxy: Optional[np.ndarray] = None
    mask: Optional[np.ndarray] = None
    confidence: Optional[np.ndarray] = None
    class_id: Optional[np.ndarray] = None
    labels: Optional[np.ndarray] = None
    tracker_id: Optional[np.ndarray] = None
    track_id: Optional[np.ndarray] = None

    def __post_init__(self):
        n = len(self.xyxy)
        # _validate_xyxy(xyxy=self.xyxy, n=n)
        _validate_mask(mask=self.mask, n=n)
        _validate_class_id(class_id=self.class_id, n=n)
        _validate_confidence(confidence=self.confidence, n=n)
        _validate_tracker_id(tracker_id=self.tracker_id, n=n)

    def __len__(self):
        """
        Returns the number of detections in the Detections object.
        """
        return len(self.xyxy)

    def __iter__(
        self,
    ) -> Iterator[
        Tuple[
            np.ndarray,
            Optional[np.ndarray],
            Optional[float],
            Optional[int],
            Optional[int],
        ]
    ]:
        """
        Iterates over the Detections object and yield a tuple of `(xyxy, mask, confidence, class_id, tracker_id)` for each detection.
        """
        for i in range(len(self.xyxy)):
            yield (
                self.xyxy[i],
                self.mask[i] if self.mask is not None else None,
                self.confidence[i] if self.confidence is not None else None,
                self.class_id[i] if self.class_id is not None else None,
                self.tracker_id[i] if self.tracker_id is not None else None,
            )

    def __eq__(self, other: Detections):
        return all(
            [
                np.array_equal(self.xyxy, other.xyxy),
                any(
                    [
                        self.mask is None and other.mask is None,
                        np.array_equal(self.mask, other.mask),
                    ]
                ),
                any(
                    [
                        self.class_id is None and other.class_id is None,
                        np.array_equal(self.class_id, other.class_id),
                    ]
                ),
                any(
                    [
                        self.confidence is None and other.confidence is None,
                        np.array_equal(self.confidence, other.confidence),
                    ]
                ),
                any(
                    [
                        self.tracker_id is None and other.tracker_id is None,
                        np.array_equal(self.tracker_id, other.tracker_id),
                    ]
                ),
            ]
        )

    @classmethod
    def from_yolov5(cls, yolov5_results) -> Detections:
        """
        Creates a Detections instance from a YOLOv5 output Detections

        Args:
            yolov5_results (yolov5.models.common.Detections): The output Detections instance from YOLOv5

        Returns:
            Detections: A new Detections object.

        Example:
            ```python
            >>> import torch
            >>> from supervision import Detections

            >>> model = torch.hub.load('ultralytics/yolov5', 'yolov5s')
            >>> results = model(IMAGE)
            >>> detections = Detections.from_yolov5(results)
            ```
        """
        yolov5_detections_predictions = yolov5_results.pred[0].cpu().cpu().numpy()
        return cls(
            xyxy=yolov5_detections_predictions[:, :4],
            confidence=yolov5_detections_predictions[:, 4],
            class_id=yolov5_detections_predictions[:, 5].astype(int),
        )

    @classmethod
    def from_yolov8(cls, yolov8_results) -> Detections:
        """
        Creates a Detections instance from a YOLOv8 output Results

        Args:
            yolov8_results (ultralytics.yolo.engine.results.Results): The output Results instance from YOLOv8

        Returns:
            Detections: A new Detections object.

        Example:
            ```python
            >>> from ultralytics import YOLO
            >>> from supervision import Detections

            >>> model = YOLO('yolov8s.pt')
            >>> yolov8_results = model(IMAGE)[0]
            >>> detections = Detections.from_yolov8(yolov8_results)
            ```
        """
        return cls(
            xyxy=yolov8_results.boxes.xyxy.cpu().numpy(),
            confidence=yolov8_results.boxes.conf.cpu().numpy(),
            class_id=yolov8_results.boxes.cls.cpu().numpy().astype(int),
        )

    @classmethod
    def from_transformers(cls, transformers_results: dict) -> Detections:
        """
        Creates a Detections instance from Object Detection Transformer output Results

        Returns:
            Detections: A new Detections object.
        """
        return cls(
            xyxy=transformers_results["boxes"].cpu().numpy(),
            confidence=transformers_results["scores"].cpu().numpy(),
            class_id=transformers_results["labels"].cpu().numpy().astype(int),
        )

    @classmethod
    def from_detectron2(cls, detectron2_results) -> Detections:
        return cls(
            xyxy=detectron2_results["instances"].pred_boxes.tensor.cpu().numpy(),
            confidence=detectron2_results["instances"].scores.cpu().numpy(),
            class_id=detectron2_results["instances"]
            .pred_classes.cpu()
            .numpy()
            .astype(int),
        )

    @classmethod
    def from_roboflow(cls, roboflow_result: dict, class_list: List[str]) -> Detections:
        xyxy = []
        confidence = []
        class_id = []

        for prediction in roboflow_result["predictions"]:
            x = prediction["x"]
            y = prediction["y"]
            width = prediction["width"]
            height = prediction["height"]
            x_min = x - width / 2
            y_min = y - height / 2
            x_max = x_min + width
            y_max = y_min + height
            xyxy.append([x_min, y_min, x_max, y_max])
            class_id.append(class_list.index(prediction["class"]))
            confidence.append(prediction["confidence"])

        return Detections(
            xyxy=np.array(xyxy),
            confidence=np.array(confidence),
            class_id=np.array(class_id).astype(int),
        )

    @classmethod
    def from_sam(cls, sam_result: List[dict]) -> Detections:
        """
        Creates a Detections instance from Segment Anything Model (SAM) by Meta AI.

        Args:
            sam_result (List[dict]): The output Results instance from SAM

        Returns:
            Detections: A new Detections object.

        Example:
            ```python
            >>> from segment_anything import sam_model_registry, SamAutomaticMaskGenerator
            >>> import supervision as sv

            >>> sam = sam_model_registry[MODEL_TYPE](checkpoint=CHECKPOINT_PATH).to(device=DEVICE)
            >>> mask_generator = SamAutomaticMaskGenerator(sam)
            >>> sam_result = mask_generator.generate(IMAGE)
            >>> detections = sv.Detections.from_sam(sam_result=sam_result)
            ```
        """
        sorted_generated_masks = sorted(
            sam_result, key=lambda x: x["area"], reverse=True
        )

        xywh = np.array([mask["bbox"] for mask in sorted_generated_masks])
        mask = np.array([mask["segmentation"] for mask in sorted_generated_masks])

        return Detections(xyxy=xywh_to_xyxy(boxes_xywh=xywh), mask=mask)

    @classmethod
    def from_coco_annotations(cls, coco_annotation: dict) -> Detections:
        xyxy, class_id = [], []

        for annotation in coco_annotation:
            x_min, y_min, width, height = annotation["bbox"]
            xyxy.append([x_min, y_min, x_min + width, y_min + height])
            class_id.append(annotation["category_id"])

        return cls(xyxy=np.array(xyxy), class_id=np.array(class_id))

    @classmethod
    def empty(cls) -> Detections:
        return cls(
            xyxy=np.empty((0, 4), dtype=np.float32),
            confidence=np.array([], dtype=np.float32),
            class_id=np.array([], dtype=int),
        )

    def gg(self):
        print("gg")
        return

    def get_anchor_coordinates(self, anchor: Position) -> np.ndarray:
        """
        Returns the bounding box coordinates for a specific anchor.

        Args:
            anchor (Position): Position of bounding box anchor for which to return the coordinates.

        Returns:
            np.ndarray: An array of shape `(n, 2)` containing the bounding box anchor coordinates in format `[x, y]`.
        """

        if anchor == Position.CENTER:
            return np.array(
                [
                    (self.xyxy[:, 0] + self.xyxy[:, 2]) / 2,
                    (self.xyxy[:, 1] + self.xyxy[:, 3]) / 2,
                ]
            ).transpose()
        elif anchor == Position.CENTER_OF_MASS:
            if self.mask is None:
                raise ValueError(
                    "Cannot use `Position.CENTER_OF_MASS` without a detection mask."
                )
            return calculate_masks_centroids(masks=self.mask)
        elif anchor == Position.CENTER_LEFT:
            return np.array(
                [
                    self.xyxy[:, 0],
                    (self.xyxy[:, 1] + self.xyxy[:, 3]) / 2,
                ]
            ).transpose()
        elif anchor == Position.CENTER_RIGHT:
            return np.array(
                [
                    self.xyxy[:, 2],
                    (self.xyxy[:, 1] + self.xyxy[:, 3]) / 2,
                ]
            ).transpose()
        elif anchor == Position.BOTTOM_CENTER:
            return np.array(
                [(self.xyxy[:, 0] + self.xyxy[:, 2]) / 2, self.xyxy[:, 3]]
            ).transpose()
        elif anchor == Position.BOTTOM_LEFT:
            return np.array([self.xyxy[:, 0], self.xyxy[:, 3]]).transpose()
        elif anchor == Position.BOTTOM_RIGHT:
            return np.array([self.xyxy[:, 2], self.xyxy[:, 3]]).transpose()
        elif anchor == Position.TOP_CENTER:
            return np.array(
                [(self.xyxy[:, 0] + self.xyxy[:, 2]) / 2, self.xyxy[:, 1]]
            ).transpose()
        elif anchor == Position.TOP_LEFT:
            return np.array([self.xyxy[:, 0], self.xyxy[:, 1]]).transpose()
        elif anchor == Position.TOP_RIGHT:
            return np.array([self.xyxy[:, 2], self.xyxy[:, 1]]).transpose()

        raise ValueError(f"{anchor} is not supported.")

    def __getitem__(self, index: np.ndarray) -> Detections:
        if isinstance(index, np.ndarray) and (
            index.dtype == bool or index.dtype == int
        ):
            return Detections(
                xyxy=self.xyxy[index],
                mask=self.mask[index] if self.mask is not None else None,
                confidence=(
                    self.confidence[index] if self.confidence is not None else None
                ),
                class_id=self.class_id[index] if self.class_id is not None else None,
                tracker_id=(
                    self.tracker_id[index] if self.tracker_id is not None else None
                ),
            )
        raise TypeError(
            f"Detections.__getitem__ not supported for index of type {type(index)}."
        )

    @property
    def area(self) -> np.ndarray:
        """
        Calculate the area of each detection in the set of object detections. If masks field is defined property
        returns are of each mask. If only box is given property return area of each box.

        Returns:
          np.ndarray: An array of floats containing the area of each detection in the format of `(area_1, area_2, ..., area_n)`, where n is the number of detections.
        """
        if self.mask is not None:
            return np.array([np.sum(mask) for mask in self.mask])
        else:
            return self.box_area

    @property
    def box_area(self) -> np.ndarray:
        """
        Calculate the area of each bounding box in the set of object detections.

        Returns:
            np.ndarray: An array of floats containing the area of each bounding box in the format of `(area_1, area_2, ..., area_n)`, where n is the number of detections.
        """
        return (self.xyxy[:, 3] - self.xyxy[:, 1]) * (self.xyxy[:, 2] - self.xyxy[:, 0])

    def with_nms(
        self, threshold: float = 0.5, class_agnostic: bool = False
    ) -> Detections:
        """
        Perform non-maximum suppression on the current set of object detections.

        Args:
            threshold (float, optional): The intersection-over-union threshold to use for non-maximum suppression. Defaults to 0.5.
            class_agnostic (bool, optional): Whether to perform class-agnostic non-maximum suppression. If True, the class_id of each detection will be ignored. Defaults to False.

        Returns:
            Detections: A new Detections object containing the subset of detections after non-maximum suppression.

        Raises:
            AssertionError: If `confidence` is None and class_agnostic is False. If `class_id` is None and class_agnostic is False.
        """
        if len(self) == 0:
            return self

        assert (
            self.confidence is not None
        ), f"Detections confidence must be given for NMS to be executed."

        if class_agnostic:
            predictions = np.hstack((self.xyxy, self.confidence.reshape(-1, 1)))
            indices = non_max_suppression(
                predictions=predictions, iou_threshold=threshold
            )
            return self[indices]

        assert self.class_id is not None, (
            f"Detections class_id must be given for NMS to be executed. If you intended to perform class agnostic "
            f"NMS set class_agnostic=True."
        )

        predictions = np.hstack(
            (self.xyxy, self.confidence.reshape(-1, 1), self.class_id.reshape(-1, 1))
        )
        indices = non_max_suppression(predictions=predictions, iou_threshold=threshold)
        return self[indices]



# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Clustering TAPIR tracks based on independent motion."""

import functools
import time
from typing import NamedTuple

from einshape import jax_einshape as einshape
import haiku as hk
import jax
import jax.numpy as jnp
import numpy as np
import optax

from juxtapose.trackers.tapnet import tapir_model


class TrainingState(NamedTuple):
    """Container for the training state."""

    params: hk.Params
    state: hk.State
    opt_state: optax.OptState
    rng: jax.Array
    step: jax.Array


def make_projection_matrix(pred_mat):
    """Convert predicted projection matrix parameters to a projection matrix."""
    pred_mat = einshape("n(coi)->ncoi", pred_mat, o=3, i=4)

    @jax.custom_vjp
    def f(x):
        return x

    def f_fwd(x):
        return f(x), tuple()

    def f_bwd(_, g):
        return (jnp.clip(g, -100, 100),)

    f.defvjp(f_fwd, f_bwd)
    pred_mat = f(pred_mat)
    orth1 = jnp.ones_like(pred_mat[..., 0:1, :-1]) * jnp.array([0.0, 0.0, 1.0])
    orth2 = pred_mat[..., 1:2, :-1] * jnp.array([1.0, 1.0, 0.0])
    orth2 = orth2 / jnp.sqrt(
        jnp.maximum(jnp.sum(jnp.square(orth2), axis=-1, keepdims=True), 1e-12)
    )
    orth3 = pred_mat[..., 2:3, :-1] * jnp.array([1.0, 1.0, 0.0])
    orth3 = orth3 - orth2 * jnp.sum(orth3 * orth2, axis=-1, keepdims=True)
    orth3 = orth3 / jnp.sqrt(
        jnp.maximum(jnp.sum(jnp.square(orth3), axis=-1, keepdims=True), 1e-12)
    )

    cross_prod = jnp.cross(orth1, orth2)
    orth3 = orth3 * jnp.sign(jnp.sum(cross_prod * orth3, axis=-1, keepdims=True))

    orth = jnp.concatenate([orth3, orth2, orth1], axis=-2)
    pred_mat = jnp.concatenate([orth, pred_mat[..., -1:]], axis=-1)
    return pred_mat


def project(pred_mat, pos_pred):
    """Project 3D points to 2D, with penalties for depth out-of-range."""
    pos_pred = jnp.concatenate([pos_pred[..., :3], pos_pred[..., 0:1] * 0 + 1], axis=-1)
    pred_pos = jnp.einsum("fcoi,nci->nfco", pred_mat, pos_pred)
    depth = jnp.minimum(2.0, jnp.maximum(pred_pos[..., 2:3] + 1.0, 0.5))
    oob = jnp.maximum(pred_pos[..., 2:3] - 2.0, 0.0) + jnp.maximum(
        0.5 - pred_pos[..., 2:3], 0.0
    )
    all_pred = pred_pos[..., 0:2] / depth
    all_pred = (
        all_pred + 0.1 * jax.random.normal(hk.next_rng_key(), shape=oob.shape) * oob
    )
    return all_pred, depth[..., 0]


def forward(
    fr_idx,
    pts_idx,
    pts,
    vis,
    num_cats=4,
    is_training=True,
    sequence_boundaries=tuple(),
):
    """Model forward pass."""

    def bn(x):
        return hk.BatchNorm(create_scale=True, create_offset=True, decay_rate=0.99)(
            x, is_training=is_training
        )

    pts_shape = pts.shape
    pts = jnp.reshape(pts * vis[..., jnp.newaxis], [pts.shape[0], -1])
    pt_state = hk.get_parameter("point_state", [pts_shape[0], 64], init=jnp.zeros)

    def centroid_init(shp, dtype):
        del shp  # unused
        centroid_weights = jax.nn.one_hot(
            jax.random.randint(hk.next_rng_key(), [384], 0, pts.shape[0]),
            pts.shape[0],
            axis=0,
        )
        centroids = jnp.transpose(centroid_weights) @ pts
        centroid_vis = jnp.transpose(centroid_weights) @ vis
        time_weight = centroid_vis
        centroids = jnp.concatenate([centroids, time_weight * 100.0], axis=1)
        centroids = jnp.transpose(centroids)
        return jnp.array(centroids, dtype=dtype)

    centroids = hk.get_parameter(
        "centroids", [pts_shape[1] * 3, 384], init=centroid_init
    )
    time_weight = jnp.abs(centroids[pts_shape[1] * 2 :, :]) / 100.0
    centroids = centroids[: pts_shape[1] * 2, :]
    vis_tile = jnp.reshape(
        jnp.tile(vis[:, :, jnp.newaxis], [1, 1, 2]), [pts.shape[0], -1]
    )
    tw_tile = jnp.reshape(
        jnp.tile(time_weight[:, jnp.newaxis, :], [1, 2, 1]), [-1, 384]
    )

    dists = jnp.square(pts * vis_tile) @ jnp.square(tw_tile)
    dists -= 2 * (pts * vis_tile) @ (centroids * tw_tile)
    dists += jnp.square(vis_tile) @ jnp.square(centroids * tw_tile)
    dists = jnp.exp(-dists * 10.0)

    dists = dists / jnp.maximum(jnp.sum(dists, axis=-1, keepdims=True), 1e-8)
    pt_state += hk.Linear(64)(dists)
    frame_state_nosmooth = hk.get_parameter(
        "frame_state",
        [pts_shape[1], 64],
        init=hk.initializers.TruncatedNormal(1.0),
    )
    conv = hk.Conv1D(64, 128, feature_group_count=64)
    frame_state = []
    for bnd in sequence_boundaries:
        frame_state.append(conv(frame_state_nosmooth[bnd[0] : bnd[1]]))
    frame_state = jnp.concatenate(frame_state, axis=0)

    frame_state = bn(frame_state)
    pt_state = bn(pt_state)
    state = jax.nn.relu(hk.Linear(64)(pt_state))
    state += hk.Linear(64)(jax.nn.relu(bn(hk.Linear(32)(state))))
    state += hk.Linear(64)(jax.nn.relu(bn(hk.Linear(32)(state))))

    truncated_normal = hk.initializers.TruncatedNormal

    base_pred = hk.get_parameter(
        "cat_pred_base",
        [3 * 64 * pts_shape[0], num_cats],
        init=truncated_normal(1.0),
    )
    fork1_pred = hk.get_parameter(
        "cat_pred_fork1",
        [3 * 64 * pts_shape[0], num_cats],
        init=lambda *args: truncated_normal(1.0)(*args) * 0.0001 + base_pred,
    )
    fork2_pred = hk.get_parameter(
        "cat_pred_fork2",
        [3 * 64 * pts_shape[0], num_cats],
        init=lambda *args: truncated_normal(1.0)(*args) * 0.0001 + base_pred,
    )

    def mul(mat):
        mat = einshape("(pio)c->pcio", mat, i=64, o=3)
        return jnp.einsum("pcio,pi->pco", mat, state) * 0.01

    pos_pred_base = mul(base_pred)[pts_idx]
    pos_pred_fork1 = mul(fork1_pred)[pts_idx]
    pos_pred_fork2 = mul(fork2_pred)[pts_idx]

    state = frame_state
    state = jax.nn.relu(hk.Linear(128)(state))
    state += hk.Linear(128)(bn(jax.nn.relu(hk.Linear(64)(state))))
    state += hk.Linear(128)(bn(jax.nn.relu(hk.Linear(64)(state))))
    state = state * 0.01

    base = hk.get_parameter(
        "mat_pred_base",
        [state.shape[-1], num_cats * 12],
        init=hk.initializers.TruncatedNormal(1.0),
    )
    fork1 = hk.get_parameter(
        "mat_pred_fork1",
        [state.shape[-1], num_cats * 12],
        init=hk.initializers.TruncatedNormal(1.0),
    )
    fork2 = hk.get_parameter(
        "mat_pred_fork2",
        [state.shape[-1], num_cats * 12],
        init=hk.initializers.TruncatedNormal(1.0),
    )
    pred_mat_base = state @ base
    pred_mat_fork1 = state @ fork1
    pred_mat_fork2 = state @ fork2

    pred_mat_base = make_projection_matrix(pred_mat_base)[fr_idx]
    pred_mat_fork1 = make_projection_matrix(pred_mat_fork1)[fr_idx]
    pred_mat_fork2 = make_projection_matrix(pred_mat_fork2)[fr_idx]

    if not is_training:
        pred_pos_all, depth_all = project(pred_mat_base, pos_pred_base)
        return pred_pos_all, depth_all
    else:
        return {
            "pos_pred_base": pos_pred_base,
            "pos_pred_fork1": pos_pred_fork1,
            "pos_pred_fork2": pos_pred_fork2,
            "pred_mat_base": pred_mat_base,
            "pred_mat_fork1": pred_mat_fork1,
            "pred_mat_fork2": pred_mat_fork2,
        }


# Create the loss.
@hk.transform_with_state
def loss_fn(
    data,
    num_cats=4,
    delete_mode=False,
    sequence_boundaries=tuple(),
    final_num_cats=28,
):
    """Computes the (scalar) LM loss on `data` w.r.t. params."""
    pts, vis, _ = data
    pts_idx = jax.random.permutation(hk.next_rng_key(), pts.shape[0])[:2048]
    fr_idx = jax.random.permutation(hk.next_rng_key(), pts.shape[1])[:1024]

    fwd = forward(
        fr_idx,
        pts_idx,
        pts,
        vis,
        num_cats=num_cats,
        sequence_boundaries=sequence_boundaries,
    )

    pts = pts[pts_idx][:, fr_idx]
    print(pts.shape)
    vis = vis[pts_idx][:, fr_idx]

    def do_fork(base, fork1, fork2, i, chunk=1):
        chunk1 = base[..., : i * chunk]
        chunk2 = fork1[..., i * chunk : ((1 + i) * chunk)]
        chunk3 = fork2[..., i * chunk : ((1 + i) * chunk)]
        chunk4 = base[..., (1 + i) * chunk :]
        return jnp.concatenate([chunk1, chunk2, chunk3, chunk4], axis=-1)

    def do_delete(base, i, chunk=1):
        chunk1 = base[..., : i * chunk]
        chunk4 = base[..., (1 + i) * chunk :]
        return jnp.concatenate([chunk1, chunk4], axis=-1)

    losses = []
    sum_vis = jnp.sum(vis)
    if delete_mode:
        all_pred, _ = project(fwd["pred_mat_base"], fwd["pos_pred_base"])
        all_err = get_err(pts, vis, all_pred)
        for i in range(fwd["pred_mat_base"].shape[-3]):
            err_i = do_delete(all_err, i)
            losses.append(loss_internal(err_i, sum_vis))
    else:
        all_pred_base, _ = project(fwd["pred_mat_base"], fwd["pos_pred_base"])
        all_err_base = get_err(pts, vis, all_pred_base)
        all_pred_fork1, _ = project(fwd["pred_mat_fork1"], fwd["pos_pred_fork1"])
        all_err_fork1 = get_err(pts, vis, all_pred_fork1)
        all_pred_fork2, _ = project(fwd["pred_mat_fork2"], fwd["pos_pred_fork2"])
        all_err_fork2 = get_err(pts, vis, all_pred_fork2)
        for i in range(fwd["pred_mat_base"].shape[-3]):
            err_i = do_fork(all_err_base, all_err_fork1, all_err_fork2, i)
            losses.append(loss_internal(err_i, sum_vis))
    if delete_mode:
        topk, _ = jax.lax.top_k(-jnp.array(losses), num_cats - final_num_cats + 3)
        accum_loss = jnp.mean(-topk)
    else:
        accum_loss = jnp.min(jnp.array(losses))
    return accum_loss, jnp.array(losses)


def huber(x):
    sqrt_x = jnp.sqrt(jnp.maximum(x, 1e-12))
    return jnp.where(x < 0.004, x, 0.004 * (2 * sqrt_x - 0.004)) * 100.0


def get_err(pts, vis, all_pred):
    tmp = pts[:, :, jnp.newaxis, :] - all_pred
    tmp = jnp.sum(jnp.square(tmp) * vis[:, :, jnp.newaxis, jnp.newaxis], axis=-1)

    return jnp.sum(tmp, axis=1)


def loss_internal(err_summed, sum_vis):
    min_loss = jnp.sum(jnp.min(err_summed, axis=1)) / sum_vis

    return min_loss


def loss_fn_wrapper(*args, **kwargs):
    (loss, aux), state = loss_fn.apply(*args, **kwargs)
    return loss, (state, aux)


def update(
    state,
    data,
    lr_mul=1.0,
    num_cats=4,
    delete_mode=False,
    sequence_boundaries=tuple(),
    optimiser=None,
    final_num_cats=28,
):
    """Does an SGD step and returns metrics."""
    rng, new_rng = jax.random.split(state.rng)
    loss_and_grad_fn = jax.value_and_grad(
        functools.partial(
            loss_fn_wrapper,
            num_cats=num_cats,
            delete_mode=delete_mode,
            sequence_boundaries=sequence_boundaries,
            final_num_cats=final_num_cats,
        ),
        has_aux=True,
    )
    (loss, (new_state, losses)), gradients = loss_and_grad_fn(
        state.params, state.state, rng, data
    )

    updates, new_opt_state = optimiser.update(gradients, state.opt_state)
    updates = jax.tree_map(lambda x: x * lr_mul, updates)
    new_params = optax.apply_updates(state.params, updates)

    new_state = TrainingState(
        params=new_params,
        state=new_state,
        opt_state=new_opt_state,
        rng=new_rng,
        step=state.step + 1,
    )

    metrics = {
        "step": state.step,
        "loss": loss,
        "losses": losses,
    }
    return new_state, metrics


@hk.transform_with_state
def forward_fn(pts_idx, pts, vis, num_cats=4, sequence_boundaries=tuple()):
    """Test-time forward function."""
    preds_all, depth_all = forward(
        jnp.arange(pts.shape[1], dtype=jnp.int32),
        pts_idx,
        pts,
        vis,
        num_cats=num_cats,
        is_training=False,
        sequence_boundaries=sequence_boundaries,
    )

    pts = pts[pts_idx]
    vis = vis[pts_idx]
    err = jnp.sum(jnp.square(pts[:, :, jnp.newaxis, :] - preds_all), axis=-1)
    return err * vis[:, :, jnp.newaxis], preds_all, depth_all


def pts_eval(
    state,
    pts_idx,
    pts,
    vis,
    num_cats=4,
    sequence_boundaries=tuple(),
):
    """Evaluate the errors for some points."""
    (err, pred_all, depth_all), _ = forward_fn.apply(
        state.params,
        state.state,
        state.rng,
        pts_idx,
        pts,
        vis,
        num_cats=num_cats,
        sequence_boundaries=sequence_boundaries,
    )

    return err, pred_all, depth_all


def init(rng, data, num_cats=1, sequence_boundaries=tuple(), optimiser=None):
    rng, init_rng = jax.random.split(rng)
    initial_params, initial_state = loss_fn.init(
        init_rng, data, num_cats=num_cats, sequence_boundaries=sequence_boundaries
    )
    initial_opt_state = optimiser.init(initial_params)
    return TrainingState(
        params=initial_params,
        state=initial_state,
        opt_state=initial_opt_state,
        rng=rng,
        step=jnp.array(0),
    )


def compute_clusters(
    separation_tracks_dict,
    separation_visibility_dict,
    demo_episode_ids,
    separation_video_shapes,
    query_features=None,
    final_num_cats=15,
    max_num_cats=25,
    low_visibility_threshold=0.1,
):
    """Compute clustering.

    Args:
      separation_tracks_dict: dict of tracks keyed by episode id, each of shape
        [num_points, num frames, 2].
      separation_visibility_dict: dict of visibility values keyed by episode id,
        each of shape [num_points, num frames].
      demo_episode_ids: demo episode ids
      separation_video_shapes: dict of video sizes (i.e. num_frames, height,
        width, channels), keyed by episode id. Currently assumes that they are all
        the same height, width.
      query_features: query features associated with each points (short ones will
        be removed)
      final_num_cats: the number of output clusters
      max_num_cats: the maximum number of clusters after splitting, before
        beginning to merge.
      low_visibility_threshold: throw out tracks with less than this fraction of
        visible frames.

    Returns:
      A dict, where low-visibility points have been removed.  "classes" is
        the class id's for remaining points, and "sum_error" is the sum of error
        for visible points.
    """
    iters_before_split = 500
    num_iters = (
        max_num_cats + (max_num_cats - final_num_cats) - 1
    ) * iters_before_split

    separation_tracks = np.concatenate(
        [separation_tracks_dict[x] for x in demo_episode_ids], axis=1
    )
    separation_visibility = np.concatenate(
        [separation_visibility_dict[x] for x in demo_episode_ids], axis=1
    )

    enough_visible = np.mean(separation_visibility, axis=-1) > low_visibility_threshold

    separation_tracks = separation_tracks[enough_visible]
    separation_visibility = separation_visibility[enough_visible]
    if query_features is not None:
        query_features = jax.tree_map(
            lambda x: x[:, enough_visible] if len(x.shape) > 1 else x,
            query_features,
        )
    separation_tracks_dict = jax.tree_map(
        lambda x: x[enough_visible], separation_tracks_dict
    )
    separation_visibility_dict = jax.tree_map(
        lambda x: x[enough_visible], separation_visibility_dict
    )

    cur = 0
    sequence_boundaries = []
    for shp in [separation_video_shapes[x] for x in demo_episode_ids]:
        sequence_boundaries.append((cur, cur + shp[0]))
        cur += shp[0]
    sequence_boundaries = tuple(sequence_boundaries)

    # Create the optimiser.
    optimiser = optax.chain(
        optax.clip_by_global_norm(1e-3),
        optax.adam(5e-2, b1=0.9, b2=0.99),
    )

    # Initialise the model parameters.
    rng = jax.random.PRNGKey(42)

    shp = separation_video_shapes[demo_episode_ids[0]]
    data = (
        jnp.array(separation_tracks / np.array([shp[2], shp[1]])),
        jnp.array(separation_visibility),
    )

    state = init(
        rng,
        data + (1,),
        num_cats=1,
        sequence_boundaries=sequence_boundaries,
        optimiser=optimiser,
    )  # +(np.zeros([100],dtype=np.int32),np.zeros([100],dtype=np.int32)))

    # Start training (note we don't include any explicit eval in this example).
    prev_time = time.time()
    log_every = 10
    num_cats = 1
    loss_curve = []
    loss_moving_average = 0
    num_since_fork = 1000
    delete_mode = False
    need_compile = True

    for step in range(num_iters):
        if step % iters_before_split == iters_before_split - 1:
            if delete_mode:
                num_cats -= 1
                to_delete = np.argmin(loss_moving_average)
                print("deleting:" + str(to_delete) + "; new num_cats:" + str(num_cats))

                def do_delete(val, chunk=1):
                    val = np.array(val)
                    lb = to_delete * chunk  # pylint: disable=cell-var-from-loop
                    ub = (to_delete + 1) * chunk  # pylint: disable=cell-var-from-loop
                    return np.concatenate([val[:, :lb], val[:, ub:]], axis=1)

                def delete_dict(param_dict):
                    param_dict["cat_pred_base"] = do_delete(
                        param_dict["cat_pred_base"]
                    )  # pylint: disable=cell-var-from-loop
                    param_dict["cat_pred_fork1"] = do_delete(
                        param_dict["cat_pred_fork1"]
                    )  # pylint: disable=cell-var-from-loop
                    param_dict["cat_pred_fork2"] = do_delete(
                        param_dict["cat_pred_fork2"]
                    )  # pylint: disable=cell-var-from-loop
                    param_dict[
                        "mat_pred_base"
                    ] = do_delete(  # pylint: disable=cell-var-from-loop
                        param_dict["mat_pred_base"], chunk=12
                    )
                    param_dict[
                        "mat_pred_fork1"
                    ] = do_delete(  # pylint: disable=cell-var-from-loop
                        param_dict["mat_pred_fork1"], chunk=12
                    )
                    param_dict[
                        "mat_pred_fork2"
                    ] = do_delete(  # pylint: disable=cell-var-from-loop
                        param_dict["mat_pred_fork2"], chunk=12
                    )

                delete_dict(state.params["~"])
                delete_dict(state.opt_state[1][0].mu["~"])
                delete_dict(state.opt_state[1][0].nu["~"])

            else:
                num_cats += 1
                to_split = jnp.argmin(loss_moving_average)
                print("splitting:" + str(to_split) + "; new num_cats:" + str(num_cats))

                def do_fork(base, fork1, fork2, chunk=1, noise=0.0, mul=1.0):
                    base = np.array(base)
                    fork1 = np.array(fork1)
                    fork2 = np.array(fork2)
                    lb = to_split * chunk  # pylint: disable=cell-var-from-loop
                    ub = (to_split + 1) * chunk  # pylint: disable=cell-var-from-loop
                    base[:, lb:ub] = fork1[:, lb:ub]
                    base = np.concatenate([base, fork2[:, lb:ub]], axis=-1)

                    def reinit(fork):
                        fork = np.concatenate(
                            [
                                fork,
                                (
                                    fork[:, lb:ub]
                                    + np.random.normal(size=[fork.shape[0], chunk])
                                    * noise
                                ),
                            ],
                            axis=-1,
                        )
                        fork[:, lb:ub] = (
                            fork[:, lb:ub]
                            + np.random.normal(size=[fork.shape[0], chunk]) * noise
                        )
                        if noise > 0:
                            fork = (
                                np.copy(base)
                                + np.random.normal(size=base.shape) * noise
                            )
                        fork = fork * mul
                        return fork

                    return base, reinit(fork1), reinit(fork2)

                def fork_dict(param_dict, noise=0.0, mul=1.0):
                    (
                        new_cpb,
                        new_cpf1,
                        new_cpf2,
                    ) = do_fork(  # pylint: disable=cell-var-from-loop
                        param_dict["cat_pred_base"],
                        param_dict["cat_pred_fork1"],
                        param_dict["cat_pred_fork2"],
                        noise=noise,
                        mul=mul,
                    )
                    param_dict["cat_pred_base"] = new_cpb
                    param_dict["cat_pred_fork1"] = new_cpf1
                    param_dict["cat_pred_fork2"] = new_cpf2
                    (
                        new_mpb,
                        new_mpf1,
                        new_mpf2,
                    ) = do_fork(  # pylint: disable=cell-var-from-loop
                        param_dict["mat_pred_base"],
                        param_dict["mat_pred_fork1"],
                        param_dict["mat_pred_fork2"],
                        chunk=12,
                        noise=noise,
                        mul=mul,
                    )
                    param_dict["mat_pred_base"] = new_mpb
                    param_dict["mat_pred_fork1"] = new_mpf1
                    param_dict["mat_pred_fork2"] = new_mpf2

                fork_dict(state.params["~"], noise=0.000001)
                fork_dict(state.opt_state[1][0].mu["~"], mul=0.0)
                fork_dict(state.opt_state[1][0].nu["~"], mul=1.0)

                state = TrainingState(
                    params=state.params,
                    state=state.state,
                    opt_state=optimiser.init(state.params),
                    rng=state.rng,
                    step=state.step,
                )

                delete_mode = num_cats == max_num_cats
            num_since_fork = 0
            loss_moving_average = 0
            need_compile = True
        if need_compile:
            update_jit = jax.jit(
                functools.partial(
                    update,
                    num_cats=num_cats,
                    delete_mode=delete_mode,
                    sequence_boundaries=sequence_boundaries,
                    optimiser=optimiser,
                    final_num_cats=final_num_cats,
                )
            )
            need_compile = False
        lr_mul = min(1.0, (num_since_fork + 1) / 20.0)

        # TODO(doersch): hardcoding the LR schedule isn't very smart
        if state.step > num_iters * 0.25:
            lr_mul = lr_mul / 2.0
        if state.step > num_iters * 0.50:
            lr_mul = lr_mul / 2.0
        if state.step > num_iters * 0.75:
            lr_mul = lr_mul / 2.0
        state, metrics = update_jit(state, data + (state.step,), lr_mul)
        loss_curve.append(metrics["loss"])
        loss_moving_average = 0.9 * loss_moving_average + 0.1 * metrics["losses"]

        if step % log_every == 0:
            steps_per_sec = log_every / (time.time() - prev_time)
            prev_time = time.time()
            metrics |= {"steps_per_sec": steps_per_sec}
            print(
                {
                    k: float(v) if k != "losses" else list(np.array(v))
                    for k, v in metrics.items()
                }
            )

        num_since_fork += 1

    pts_eval_jit = jax.jit(
        functools.partial(
            pts_eval,
            num_cats=num_cats,
            sequence_boundaries=sequence_boundaries,
        )
    )
    sum_error = []

    for i in range(0, separation_tracks.shape[0], 128):
        err, _, _ = pts_eval_jit(
            state,
            np.arange(i, min(separation_tracks.shape[0], i + 128)),
            data[0],
            data[1],
        )
        sum_error.append(np.sum(err, axis=1))

    sum_error = np.concatenate(sum_error, axis=0)

    return {
        "classes": np.array(np.argmin(np.array(sum_error), axis=-1)),
        "sum_error": sum_error,
        "separation_visibility": separation_visibility_dict,
        "separation_tracks": separation_tracks_dict,
        "query_features": query_features,
        "demo_episode_ids": demo_episode_ids,
    }


def construct_fake_causal_state(
    query_features, convert_to_jax=False, channel_multiplier=4
):
    """Constructs a fake TAPIR causal state which can be used to reduce jitting.

    Please not this function is very fragile and only works for the current
    version of tapir. It will likely need to be updated if TAPIR changes, but it
    is helpful for quick iterations.

    Args:
      query_features: Query features which will be used to infer shapes.
      convert_to_jax: Whether to convert it to a jax array (helps prevent
        recompiles)
      channel_multiplier: for compatibility with smaller models

    Returns:
      A causal state.
    """
    num_points = query_features_count(query_features)
    num_resolutions = len(query_features.resolutions)
    dims = 512 * channel_multiplier

    value_shapes = {
        "tapir/~/pips_mlp_mixer/block_1_causal_1": (1, num_points, 2, 512),
        "tapir/~/pips_mlp_mixer/block_1_causal_2": (1, num_points, 2, dims),
        "tapir/~/pips_mlp_mixer/block_2_causal_1": (1, num_points, 2, 512),
        "tapir/~/pips_mlp_mixer/block_2_causal_2": (1, num_points, 2, dims),
        "tapir/~/pips_mlp_mixer/block_3_causal_1": (1, num_points, 2, 512),
        "tapir/~/pips_mlp_mixer/block_3_causal_2": (1, num_points, 2, dims),
        "tapir/~/pips_mlp_mixer/block_4_causal_1": (1, num_points, 2, 512),
        "tapir/~/pips_mlp_mixer/block_4_causal_2": (1, num_points, 2, dims),
        "tapir/~/pips_mlp_mixer/block_5_causal_1": (1, num_points, 2, 512),
        "tapir/~/pips_mlp_mixer/block_5_causal_2": (1, num_points, 2, dims),
        "tapir/~/pips_mlp_mixer/block_6_causal_1": (1, num_points, 2, 512),
        "tapir/~/pips_mlp_mixer/block_6_causal_2": (1, num_points, 2, dims),
        "tapir/~/pips_mlp_mixer/block_7_causal_1": (1, num_points, 2, 512),
        "tapir/~/pips_mlp_mixer/block_7_causal_2": (1, num_points, 2, dims),
        "tapir/~/pips_mlp_mixer/block_8_causal_1": (1, num_points, 2, 512),
        "tapir/~/pips_mlp_mixer/block_8_causal_2": (1, num_points, 2, dims),
        "tapir/~/pips_mlp_mixer/block_9_causal_1": (1, num_points, 2, 512),
        "tapir/~/pips_mlp_mixer/block_9_causal_2": (1, num_points, 2, dims),
        "tapir/~/pips_mlp_mixer/block_10_causal_1": (1, num_points, 2, 512),
        "tapir/~/pips_mlp_mixer/block_10_causal_2": (1, num_points, 2, dims),
        "tapir/~/pips_mlp_mixer/block_11_causal_1": (1, num_points, 2, 512),
        "tapir/~/pips_mlp_mixer/block_11_causal_2": (1, num_points, 2, dims),
        "tapir/~/pips_mlp_mixer/block_causal_1": (1, num_points, 2, 512),
        "tapir/~/pips_mlp_mixer/block_causal_2": (1, num_points, 2, dims),
    }
    fake_ret = {k: np.zeros(v) for k, v in value_shapes.items()}
    fake_ret = [fake_ret] * num_resolutions * 4
    if convert_to_jax:
        fake_ret = jax.tree_map(jnp.array, fake_ret)
    return fake_ret


def _build_online_model_init(frames, query_points, *, tapir_model_kwargs):
    """Build tapir model for initialisation and tracking.

    Args:
      frames:
      query_points:
      tapir_model_kwargs:

    Returns:

    Raises:
      <Any>:
    """
    if not tapir_model_kwargs["use_causal_conv"]:
        raise ValueError("Online model requires causal TAPIR training.")
    model = tapir_model.TAPIR(**tapir_model_kwargs)
    feature_grids = model.get_feature_grids(
        frames,
        is_training=False,
    )
    query_features = model.get_query_features(
        frames,
        is_training=True,
        query_points=query_points,
        feature_grids=feature_grids,
    )
    return query_features


def _build_online_model_predict(
    frames,
    query_features,
    causal_context,
    *,
    tapir_model_kwargs,
    query_chunk_size=256,
    query_points_in_video=None,
):
    """Compute point tracks and occlusions given frames and query points."""
    if not tapir_model_kwargs["use_causal_conv"]:
        raise ValueError("Online model requires causal TAPIR training.")
    model = tapir_model.TAPIR(**tapir_model_kwargs)
    feature_grids = model.get_feature_grids(
        frames,
        is_training=False,
    )
    trajectories = model.estimate_trajectories(
        frames.shape[-3:-1],
        is_training=False,
        feature_grids=feature_grids,
        query_features=query_features,
        query_points_in_video=query_points_in_video,
        query_chunk_size=query_chunk_size,
        causal_context=causal_context,
        get_causal_context=True,
    )
    trajectories = dict(trajectories)
    causal_context = trajectories["causal_context"]
    del trajectories["causal_context"]
    return {k: v[-1] for k, v in trajectories.items()}, causal_context


def build_models(
    checkpoint_path,
    query_chunk_size=256,
):
    """Build tapir model for initialisation and tracking."""
    ckpt_state = np.load(checkpoint_path, allow_pickle=True).item()
    params, state = ckpt_state["params"], ckpt_state["state"]

    num_params = hk.data_structures.tree_size(params)
    num_bytes = hk.data_structures.tree_bytes(params)
    print("TAPIR model")
    print(
        f"Number of params: {num_params}",
    )
    print(f"Number of bytes: {num_bytes / 1e6:.2f} MB")

    tapir_model_kwargs = dict(
        use_causal_conv=True, bilinear_interp_with_depthwise_conv=False
    )

    online_model_init_fn = functools.partial(
        _build_online_model_init,
        tapir_model_kwargs=tapir_model_kwargs,
    )
    online_init = hk.transform_with_state(online_model_init_fn)
    online_init_apply = jax.jit(online_init.apply, backend="cpu")
    online_model_predict_fn = functools.partial(
        _build_online_model_predict,
        tapir_model_kwargs=tapir_model_kwargs,
        query_chunk_size=query_chunk_size,
    )
    online_predict = hk.transform_with_state(online_model_predict_fn)
    # Jit is broken here unfortunately.
    online_predict_apply_cpu = online_predict.apply
    online_predict_apply_gpu = jax.jit(online_predict.apply, backend="gpu")
    # online_predict_apply = online_predict.apply

    rng = jax.random.PRNGKey(42)
    online_init_apply = functools.partial(
        online_init_apply, params=params, state=state, rng=rng
    )
    online_predict_apply_cpu = functools.partial(
        online_predict_apply_cpu, params=params, state=state, rng=rng
    )
    online_predict_apply_gpu = functools.partial(
        online_predict_apply_gpu, params=params, state=state, rng=rng
    )
    return online_init_apply, online_predict_apply_cpu, online_predict_apply_gpu


def query_features_join(feature_list):
    """Merge a list of query features int a single query features structure."""
    lowres = [x.lowres for x in feature_list]
    hires = [x.hires for x in feature_list]

    joined_features = tapir_model.QueryFeatures(
        lowres=tuple(np.concatenate(x, axis=1) for x in zip(*lowres)),
        hires=tuple(np.concatenate(x, axis=1) for x in zip(*hires)),
        resolutions=feature_list[0].resolutions,
    )
    return joined_features


def query_features_count(features):
    """Number of points within a query features structure."""
    return features.lowres[0].shape[1]


def predictions_to_tracks_visibility(predictions, single_step=True):
    """Extract tracks and visibility from TAPIR predictions.

    Args:
      predictions: Predictions output of TAPIR
      single_step: Whether we are processing a single step or a whole episode.

    Returns:
      * Tracks of shape [num_points, (t), xy]
      * Visibility of shape [num_points, (t)]. Float between 0 and 1.
    """
    tracks = predictions["tracks"][0]
    occlusion = predictions["occlusion"][0]
    expected_dist = predictions["expected_dist"][0]
    if single_step:
        tracks = tracks[:, 0]
        occlusion = occlusion[:, 0]
        expected_dist = expected_dist[:, 0]
    pred_occ = jax.nn.sigmoid(occlusion)
    visibility = (1 - pred_occ) * (1 - jax.nn.sigmoid(expected_dist))
    return tracks, visibility


def preprocess_frames(frames: np.ndarray) -> np.ndarray:
    """Preprocess frames to model inputs.

    Args:
      frames: [num_frames, height, width, 3], [0, 255], np.uint8

    Returns:
      frames: [num_frames, height, width, 3], [-1, 1], np.float32
    """
    frames = frames.astype(np.float32)
    frames = frames / 255 * 2 - 1
    return frames


def track_many_points(
    separation_videos,
    demo_episode_ids,
    checkpoint_path,
    frame_stride=4,
    points_per_frame=8,
    point_batch_size=2048,
    sample_box_corners=(0.1, 0.1, 0.9, 0.9),
):
    """Track random points sampled from the videos.

    Args:
      separation_videos: dict of uint8 tensors, keyed by demo_episode_ids.
      demo_episode_ids: demo episode ids.
      checkpoint_path: path to TAPIR checkpoint file
      frame_stride: sample points from one out of every this number of frames
      points_per_frame: from each selected frame, sample this many points
      point_batch_size: how many points to run through tapir simultaneously.
      sample_box_corners: sample within this box.  Format is [x_lower, y_lower,
        x_upper, y_upper], in normalized coordinates.

    Returns:
      a dict with tracked points
    """
    tapir_init, _, tapir_predict = build_models(checkpoint_path, query_chunk_size=512)
    np.random.seed(42)

    query_features = []
    query_features2 = []
    query_points = []
    tmp_query_points = []

    def merge_struct(query_features, tmp_query_points):
        query_features2.append(query_features_join(query_features))
        query_points.append(
            [np.concatenate([x[i] for x in tmp_query_points], axis=0) for i in range(3)]
        )

    for sv_idx, sv in enumerate([separation_videos[x] for x in demo_episode_ids]):
        print(f"extracting query features for video {sv_idx}")
        for i in range(0, len(sv), frame_stride):
            x_scl = sample_box_corners[2] - sample_box_corners[0]
            y_scl = sample_box_corners[3] - sample_box_corners[1]
            x_add = sample_box_corners[0]
            y_add = sample_box_corners[1]
            qp = (
                np.random.uniform(0.0, 1.0, [points_per_frame, 3])
                * np.array([0.0, sv.shape[1] * y_scl, sv.shape[2] * x_scl])[None, ...]
                + np.array([0.0, sv.shape[1] * y_add, sv.shape[2] * x_add])[None, ...]
            )
            tmp_query_points.append(
                (
                    np.array([sv_idx] * points_per_frame),
                    np.array([i] * points_per_frame),
                    qp[..., 1:],
                )
            )

            qf, _ = tapir_init(
                frames=preprocess_frames(sv[None, None, i]),
                query_points=qp[None],
            )
            query_features.append(qf)
            if len(query_features) == point_batch_size // points_per_frame:
                merge_struct(query_features, tmp_query_points)
                query_features = []
                tmp_query_points = []

    print("Done extracting query features")

    num_extra = 0
    if query_features:
        merge_struct(query_features, tmp_query_points)
    out_query_features = query_features_join(query_features2)
    out_query_points = [
        np.concatenate([x[i] for x in query_points], axis=0) for i in range(3)
    ]

    if query_features:
        del query_features2[-1]
        del query_points[-1]
        while len(query_features) < point_batch_size // points_per_frame:
            query_features.append(query_features[-1])
            tmp_query_points.append(tmp_query_points[-1])
            num_extra += points_per_frame
        merge_struct(query_features, tmp_query_points)

    all_separation_tracks = []
    all_separation_visibility = []

    print("Note that TAPIR compilation takes a while.")
    print("You may not see GPU usage immediately.")
    for qf_idx, query_features in enumerate(query_features2):
        print(f"query feature batch {qf_idx}/{len(query_features2)}")
        separation_tracks = []
        separation_visibility = []
        for sv_idx, sv in enumerate([separation_videos[x] for x in demo_episode_ids]):
            print(f"tracking in video {sv_idx}/{len(demo_episode_ids)}")
            causal_state = construct_fake_causal_state(
                query_features,
                convert_to_jax=True,
                channel_multiplier=4,
            )
            for i in range(len(sv)):
                (prediction, causal_state), _ = tapir_predict(
                    frames=preprocess_frames(sv[None, None, i]),
                    query_features=query_features,
                    causal_context=causal_state,
                )

                prediction = jax.tree_map(np.array, prediction)

                res = predictions_to_tracks_visibility(prediction)
                separation_tracks.append(res[0])
                separation_visibility.append(res[1] > 0.5)  # Threshold
        all_separation_visibility.append(np.stack(separation_visibility, axis=1))
        all_separation_tracks.append(np.stack(separation_tracks, axis=1))

    separation_visibility = np.concatenate(all_separation_visibility, axis=0)
    separation_tracks = np.concatenate(all_separation_tracks, axis=0)

    pad_start = separation_tracks.shape[0] - num_extra
    separation_tracks = separation_tracks[:pad_start]
    separation_visibility = separation_visibility[:pad_start]

    separation_video_shapes = [separation_videos[x].shape for x in demo_episode_ids]
    bnds = []
    cur = 0
    for shp in separation_video_shapes:
        bnds.append((cur, cur + shp[0]))
        cur += shp[0]

    separation_visibility = {
        k: separation_visibility[:, lb:ub]
        for k, (lb, ub) in zip(demo_episode_ids, bnds)
    }
    separation_tracks = {
        k: separation_tracks[:, lb:ub] for k, (lb, ub) in zip(demo_episode_ids, bnds)
    }
    return {
        "separation_visibility": separation_visibility,
        "separation_tracks": separation_tracks,
        "video_shape": {
            x: separation_video_shapes[i] for i, x in enumerate(demo_episode_ids)
        },
        "query_features": jax.tree_map(np.array, out_query_features),
        "demo_episode_ids": demo_episode_ids,
        "query_points": out_query_points,
    }



# https://github.com/open-mmlab/mmdeploy/blob/main/docs/en/04-supported-codebases/mmpose.md

from typing import List, Tuple
import numpy as np
from pathlib import Path
import onnxruntime as ort

from .postprocessing import get_simcc_maximum
from .preprocessing import bbox_xyxy2cs, top_down_affine
from juxtapose.utils import LOGGER
from juxtapose.utils.downloads import safe_download

base = "juxtapose"


class RTMPose:
    """RTMPose model (s, m, l) to detect multi-person poses (class 0) based on bboxes"""

    def __init__(self, type: str = "m", device: str = "cpu") -> None:
        download_dir = Path("model")

        onnx_model = download_dir / f"rtmpose-{type}.onnx"

        if not onnx_model.exists():
            safe_download(
                f"https://huggingface.co/ziq/rtm/resolve/main/rtmpose-{type}.onnx",
                file=f"rtmpose-{type}",
                dir=download_dir,
            )

        providers = {"cpu": "CPUExecutionProvider", "cuda": "CUDAExecutionProvider"}[
            device
        ]

        self.session = ort.InferenceSession(
            path_or_bytes=onnx_model, providers=[providers]
        )

        self.onnx_model = onnx_model
        self.model_input_size = (192, 256)
        self.mean = (123.675, 116.28, 103.53)
        self.std = (58.395, 57.12, 57.375)
        self.device = device
        LOGGER.info(f"Loaded rtmpose-{type} onnx model.")
        # self.conf_thres = conf_thres

    def __call__(self, im: np.ndarray, bboxes: list = []):
        """Return List of 17 xy coordinates based on im frame (cv2.imread)
        im -> (h, w, c)
        bboxes -> [[x, y, x, y], [x, y, x, y], ...] -> two persons detected
        return -> np.ndarray([[x, y, ... 17 times], [x, y, ... 17 times], ...]) -> (2 or more, 17, 2)
        """

        if len(bboxes) == 0:
            bboxes = [[0, 0, im.shape[1], im.shape[0]]]

        keypoints, scores = [], []
        for bbox in bboxes:
            img, center, scale = self.preprocess(im, bbox)
            outputs = self.inference(img)
            kpts, score = self.postprocess(outputs, center, scale)

            keypoints.append(kpts)
            scores.append(score)

        keypoints = np.concatenate(keypoints, axis=0)
        scores = np.concatenate(scores, axis=0)

        return keypoints, scores

    def inference(self, im: np.array):
        im = im.transpose(2, 0, 1)
        im = np.ascontiguousarray(im, dtype=np.float32)
        input = im[None, :, :, :]

        sess_input = {self.session.get_inputs()[0].name: input}
        sess_output = []
        for out in self.session.get_outputs():
            sess_output.append(out.name)

        outputs = self.session.run(sess_output, sess_input)
        return outputs

    def preprocess(self, img: np.ndarray, bbox: list):
        """Do preprocessing for RTMPose model inference.

        Args:
            img (np.ndarray): Input image in shape.
            bbox (list):  xyxy-format bounding box of target.

        Returns:
            tuple:
            - resized_img (np.ndarray): Preprocessed image.
            - center (np.ndarray): Center of image.
            - scale (np.ndarray): Scale of image.
        """
        bbox = np.array(bbox)

        # get center and scale
        center, scale = bbox_xyxy2cs(bbox, padding=1.25)

        # do affine transformation
        resized_img, scale = top_down_affine(self.model_input_size, scale, center, img)
        # normalize image
        if self.mean is not None:
            self.mean = np.array(self.mean)
            self.std = np.array(self.std)
            resized_img = (resized_img - self.mean) / self.std

        return resized_img, center, scale

    def postprocess(
        self,
        outputs: List[np.ndarray],
        center: Tuple[int, int],
        scale: Tuple[int, int],
        simcc_split_ratio: float = 2.0,
    ) -> Tuple[np.ndarray, np.ndarray]:
        """Postprocess for RTMPose model output.

        Args:
            outputs (np.ndarray): Output of RTMPose model.
            model_input_size (tuple): RTMPose model Input image size.
            center (tuple): Center of bbox in shape (x, y).
            scale (tuple): Scale of bbox in shape (w, h).
            simcc_split_ratio (float): Split ratio of simcc.

        Returns:
            tuple:
            - keypoints (np.ndarray): Rescaled keypoints.
            - scores (np.ndarray): Model predict scores.
        """
        # decode simcc
        simcc_x, simcc_y = outputs
        locs, scores = get_simcc_maximum(simcc_x, simcc_y)
        keypoints = locs / simcc_split_ratio

        # rescale keypoints
        keypoints = keypoints / self.model_input_size * scale
        keypoints = keypoints + center - scale / 2

        return keypoints, scores



import torch

from juxtapose.utils.checks import check_version


TORCH_1_9 = check_version(torch.__version__, "1.9.0")
TORCH_1_11 = check_version(torch.__version__, "1.11.0")
TORCH_1_12 = check_version(torch.__version__, "1.12.0")
TORCH_2_0 = check_version(torch.__version__, "2.0.0")


def smart_inference_mode(fn):
    """Applies torch.inference_mode() decorator if torch>=1.9.0 else torch.no_grad() decorator."""

    def decorate(*args, **kwargs):
        """Applies appropriate torch decorator for inference mode based on torch version."""
        return (torch.inference_mode if TORCH_1_9 else torch.no_grad)()(fn)(
            *args, **kwargs
        )

    return decorate



# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Python script to generate a pickle with annotations from raw videos."""

import csv
import dataclasses
import io
import math
import os
import pickle
from typing import Dict, Iterator, List, Sequence, Tuple

from absl import app
from absl import flags
from absl import logging
import ffmpeg
import numpy as np
from PIL import Image

FLAGS = flags.FLAGS

flags.DEFINE_string('input_csv_path', None, 'Path to the input csv.')
flags.DEFINE_string(
    'output_base_path',
    None,
    'Path to the output folder where pickle files will be stored.',
)
flags.DEFINE_string(
    'video_root_path',
    None,
    (
        'Path to the root directory of the extracted Kinetics-700-2020 videos'
        ' from the validation set.'
    ),
)
flags.DEFINE_integer('num_shards', 10, 'Number of pickle shards to output.')

_JPEG_HEADER = b'\xff\xd8'


@dataclasses.dataclass(frozen=True)
class Point:
  x: float
  y: float
  occluded: bool


@dataclasses.dataclass(frozen=True)
class Track:
  points: Tuple[Point, ...]


@dataclasses.dataclass(frozen=True)
class Video:
  youtube_id: str
  start_time_sec: int
  end_time_sec: int
  video_path: str
  tracks: Tuple[Track, ...]


def csv_to_dataset(
    csv_path: str, videos_path: Dict[str, str]
) -> Tuple[Video, ...]:
  """Reads the input CSV and creates a list of `Video`s out of that."""

  def points(row: Sequence[str]) -> Iterator[Point]:
    for i in range(250):
      x, y, occ = row[3 + 3 * i : 3 + 3 * i + 3]
      x = float(x)
      y = float(y)
      assert occ in ('0', '1')
      occ = occ == '1'
      yield Point(x, y, occ)

  logging.info('Reading CSV "%s".', csv_path)

  with open(csv_path) as f:
    reader = csv.reader(f, delimiter=',')

    tracks_per_video: Dict[Tuple[str, int, int], List[Track]] = {}
    for row in reader:
      assert len(row) == 3 + 3 * 250

      youtube_id, start_time_sec, end_time_sec = row[:3]
      start_time_sec = int(start_time_sec)
      end_time_sec = int(end_time_sec)
      key = (youtube_id, start_time_sec, end_time_sec)

      track = Track(tuple(points(row)))

      if key not in tracks_per_video:
        tracks_per_video[key] = []
      tracks_per_video[key].append(track)

    def videos() -> Iterator[Video]:
      for key, tracks in tracks_per_video.items():
        youtube_id, start_time_sec, end_time_sec = key

        name = f'{youtube_id}_{start_time_sec:06}_{end_time_sec:06}'
        if name not in videos_path:
          logging.warning('Video "%s" not downloaded. Skipping it.', name)
          continue
        video_path = videos_path[name]

        yield Video(
            youtube_id, start_time_sec, end_time_sec, video_path, tuple(tracks)
        )

    return tuple(videos())


def get_paths_to_videos(video_root_path: str) -> Dict[str, str]:
  """Returns the relative path to each downloaded video."""
  logging.info('Reading all videos in subfolders of "%s".', video_root_path)
  video_to_path: Dict[str, str] = {}
  for folder_or_video in os.listdir(video_root_path):
    path = os.path.join(video_root_path, folder_or_video)
    if os.path.isdir(path):
      subfolder_paths = get_paths_to_videos(path)
      for k, v in subfolder_paths.items():
        assert k not in video_to_path
        video_to_path[k] = v
    elif folder_or_video.endswith('.mp4'):
      name = folder_or_video[:-4]  # Remove '.mp4'.
      assert name not in video_to_path
      video_to_path[name] = path

  return video_to_path


def extract_frames(video_path: str, fps: float) -> Tuple[bytes, ...]:
  """Extracts list of jpeg bytes from the given video using ffmpeg."""
  cmd = (
      ffmpeg.input(video_path)
      .filter('fps', fps=fps)
      .output('pipe:', format='image2pipe')
  )
  jpeg_bytes, _ = cmd.run(capture_stdout=True, quiet=True)
  jpeg_bytes = jpeg_bytes.split(_JPEG_HEADER)[1:]
  jpeg_bytes = map(lambda x: _JPEG_HEADER + x, jpeg_bytes)
  return tuple(jpeg_bytes)


def generate_example(video: Video) -> Dict[str, np.ndarray]:
  """Generates a dictionary with the info from a `Video`."""
  example: Dict[str, np.ndarray] = {}

  imgs_encoded = extract_frames(video.video_path, 25.0)
  if len(imgs_encoded) > 250:
    imgs_encoded = imgs_encoded[:250]

  if len(imgs_encoded) < 250:
    # Clip is shorter than 10s.
    num_frames = len(imgs_encoded)
    new_tracks = tuple(
        Track(tuple(t.points[:num_frames])) for t in video.tracks
    )
    video = Video(
        video.youtube_id,
        video.start_time_sec,
        video.end_time_sec,
        video.video_path,
        new_tracks,
    )

  example['video'] = np.array(imgs_encoded)
  byteio = io.BytesIO(imgs_encoded[0])
  img = Image.open(byteio)
  height, width, _ = np.array(img).shape

  points = []
  occluded = []
  for track in video.tracks:
    points.append(
        [
            [(p.x * width - 0.5) / width, (p.y * height - 0.5) / height]
            for p in track.points
        ]
    )
    occluded.append([p.occluded for p in track.points])

  example['points'] = np.array(points, dtype=np.float64)
  example['occluded'] = np.array(occluded, dtype=bool)

  return example


def main(argv: Sequence[str]) -> None:
  del argv

  output_folder = FLAGS.output_base_path
  if output_folder and not os.path.exists(output_folder):
    os.makedirs(output_folder)

  # Reads data.
  videos_path = get_paths_to_videos(FLAGS.video_root_path)
  videos = csv_to_dataset(FLAGS.input_csv_path, videos_path)

  # Process the dataset and store pickles.
  num_examples_per_shard = int(math.ceil(len(videos) / FLAGS.num_shards))
  shard = 0
  data = []
  for i, video in enumerate(videos):
    print(
        'Processing example %d of %d   (%d%%) \r'
        % (i, len(videos), i * 100 / len(videos)),
        end='',
    )
    data.append(generate_example(video))
    if i == len(videos) - 1 or len(data) == num_examples_per_shard:
      shard_path = os.path.join(
          output_folder, f'{shard:04}_of_{FLAGS.num_shards:04}.pkl'
      )
      logging.info('Writing file "%s".', shard_path)
      with open(shard_path, 'wb') as f:
        pickle.dump(data, f)
      data.clear()
      shard += 1


if __name__ == '__main__':
  app.run(main)



import contextlib
import time
import re

import torch
import numpy as np


class Profile(contextlib.ContextDecorator):
    """
    POSE Profile class.
    Usage: as a decorator with @Profile() or as a context manager with 'with Profile():'
    """

    def __init__(self, t=0.0):
        """
        Initialize the Profile class.

        Args:
            t (float): Initial time. Defaults to 0.0.
        """
        self.t = t
        self.cuda = torch.cuda.is_available()

    def __enter__(self):
        """
        Start timing.
        """
        self.start = self.time()
        return self

    def __exit__(self, type, value, traceback):
        """
        Stop timing.
        """
        self.dt = self.time() - self.start  # delta-time
        self.t += self.dt  # accumulate dt

    def time(self):
        """
        Get current time.
        """
        if self.cuda:
            torch.cuda.synchronize()
        return time.time()


def clean_str(s):
    """
    Cleans a string by replacing special characters with underscore _

    Args:
      s (str): a string needing special characters replaced

    Returns:
      (str): a string with special characters replaced by an underscore _
    """
    return re.sub(pattern="[|@#!¡·$€%&()=?¿^*;:,¨´><+]", repl="_", string=s)


def xyxy2xyxyxyxy(xyxy):
    """
    xyxy: [(x1, y1, x2, y2), ...]
    """
    xyxyxyxy = []
    for _xyxy in xyxy:
        x1, y1, x2, y2 = _xyxy

        x_min = min(x1, x2)
        y_min = min(y1, y2)
        x_max = max(x1, x2)
        y_max = max(y1, y2)

        xyxyxyxy.append(
            np.array([(x_min, y_min), (x_min, y_max), (x_max, y_max), (x_max, y_min)])
        )
    return xyxyxyxy



# ------------------------------------------------------------------------
# Grounding DINO
# url: https://github.com/IDEA-Research/GroundingDINO
# Copyright (c) 2023 IDEA. All Rights Reserved.
# Licensed under the Apache License, Version 2.0 [see LICENSE for details]
# ------------------------------------------------------------------------

import torch
import torch.nn as nn
import torch.nn.functional as F

# from timm.models.layers import DropPath


def drop_path(
    x, drop_prob: float = 0.0, training: bool = False, scale_by_keep: bool = True
):
    """Drop paths (Stochastic Depth) per sample (when applied in main path of residual blocks).

    This is the same as the DropConnect impl I created for EfficientNet, etc networks, however,
    the original name is misleading as 'Drop Connect' is a different form of dropout in a separate paper...
    See discussion: https://github.com/tensorflow/tpu/issues/494#issuecomment-532968956 ... I've opted for
    changing the layer and argument names to 'drop path' rather than mix DropConnect as a layer name and use
    'survival rate' as the argument.

    """
    if drop_prob == 0.0 or not training:
        return x
    keep_prob = 1 - drop_prob
    shape = (x.shape[0],) + (1,) * (
        x.ndim - 1
    )  # work with diff dim tensors, not just 2D ConvNets
    random_tensor = x.new_empty(shape).bernoulli_(keep_prob)
    if keep_prob > 0.0 and scale_by_keep:
        random_tensor.div_(keep_prob)
    return x * random_tensor


class DropPath(nn.Module):
    """Drop paths (Stochastic Depth) per sample  (when applied in main path of residual blocks)."""

    def __init__(self, drop_prob: float = 0.0, scale_by_keep: bool = True):
        super(DropPath, self).__init__()
        self.drop_prob = drop_prob
        self.scale_by_keep = scale_by_keep

    def forward(self, x):
        return drop_path(x, self.drop_prob, self.training, self.scale_by_keep)

    def extra_repr(self):
        return f"drop_prob={round(self.drop_prob,3):0.3f}"


class FeatureResizer(nn.Module):
    """
    This class takes as input a set of embeddings of dimension C1 and outputs a set of
    embedding of dimension C2, after a linear transformation, dropout and normalization (LN).
    """

    def __init__(self, input_feat_size, output_feat_size, dropout, do_ln=True):
        super().__init__()
        self.do_ln = do_ln
        # Object feature encoding
        self.fc = nn.Linear(input_feat_size, output_feat_size, bias=True)
        self.layer_norm = nn.LayerNorm(output_feat_size, eps=1e-12)
        self.dropout = nn.Dropout(dropout)

    def forward(self, encoder_features):
        x = self.fc(encoder_features)
        if self.do_ln:
            x = self.layer_norm(x)
        output = self.dropout(x)
        return output


def l1norm(X, dim, eps=1e-8):
    """L1-normalize columns of X"""
    norm = torch.abs(X).sum(dim=dim, keepdim=True) + eps
    X = torch.div(X, norm)
    return X


def l2norm(X, dim, eps=1e-8):
    """L2-normalize columns of X"""
    norm = torch.pow(X, 2).sum(dim=dim, keepdim=True).sqrt() + eps
    X = torch.div(X, norm)
    return X


def func_attention(query, context, smooth=1, raw_feature_norm="softmax", eps=1e-8):
    """
    query: (n_context, queryL, d)
    context: (n_context, sourceL, d)
    """
    batch_size_q, queryL = query.size(0), query.size(1)
    batch_size, sourceL = context.size(0), context.size(1)

    # Get attention
    # --> (batch, d, queryL)
    queryT = torch.transpose(query, 1, 2)

    # (batch, sourceL, d)(batch, d, queryL)
    # --> (batch, sourceL, queryL)
    attn = torch.bmm(context, queryT)
    if raw_feature_norm == "softmax":
        # --> (batch*sourceL, queryL)
        attn = attn.view(batch_size * sourceL, queryL)
        attn = nn.Softmax()(attn)
        # --> (batch, sourceL, queryL)
        attn = attn.view(batch_size, sourceL, queryL)
    elif raw_feature_norm == "l2norm":
        attn = l2norm(attn, 2)
    elif raw_feature_norm == "clipped_l2norm":
        attn = nn.LeakyReLU(0.1)(attn)
        attn = l2norm(attn, 2)
    else:
        raise ValueError("unknown first norm type:", raw_feature_norm)
    # --> (batch, queryL, sourceL)
    attn = torch.transpose(attn, 1, 2).contiguous()
    # --> (batch*queryL, sourceL)
    attn = attn.view(batch_size * queryL, sourceL)
    attn = nn.Softmax()(attn * smooth)
    # --> (batch, queryL, sourceL)
    attn = attn.view(batch_size, queryL, sourceL)
    # --> (batch, sourceL, queryL)
    attnT = torch.transpose(attn, 1, 2).contiguous()

    # --> (batch, d, sourceL)
    contextT = torch.transpose(context, 1, 2)
    # (batch x d x sourceL)(batch x sourceL x queryL)
    # --> (batch, d, queryL)
    weightedContext = torch.bmm(contextT, attnT)
    # --> (batch, queryL, d)
    weightedContext = torch.transpose(weightedContext, 1, 2)

    return weightedContext, attnT


class BiMultiHeadAttention(nn.Module):
    def __init__(self, v_dim, l_dim, embed_dim, num_heads, dropout=0.1, cfg=None):
        super(BiMultiHeadAttention, self).__init__()

        self.embed_dim = embed_dim
        self.num_heads = num_heads
        self.head_dim = embed_dim // num_heads
        self.v_dim = v_dim
        self.l_dim = l_dim

        assert (
            self.head_dim * self.num_heads == self.embed_dim
        ), f"embed_dim must be divisible by num_heads (got `embed_dim`: {self.embed_dim} and `num_heads`: {self.num_heads})."
        self.scale = self.head_dim ** (-0.5)
        self.dropout = dropout

        self.v_proj = nn.Linear(self.v_dim, self.embed_dim)
        self.l_proj = nn.Linear(self.l_dim, self.embed_dim)
        self.values_v_proj = nn.Linear(self.v_dim, self.embed_dim)
        self.values_l_proj = nn.Linear(self.l_dim, self.embed_dim)

        self.out_v_proj = nn.Linear(self.embed_dim, self.v_dim)
        self.out_l_proj = nn.Linear(self.embed_dim, self.l_dim)

        self.stable_softmax_2d = True
        self.clamp_min_for_underflow = True
        self.clamp_max_for_overflow = True

        self._reset_parameters()

    def _shape(self, tensor: torch.Tensor, seq_len: int, bsz: int):
        return (
            tensor.view(bsz, seq_len, self.num_heads, self.head_dim)
            .transpose(1, 2)
            .contiguous()
        )

    def _reset_parameters(self):
        nn.init.xavier_uniform_(self.v_proj.weight)
        self.v_proj.bias.data.fill_(0)
        nn.init.xavier_uniform_(self.l_proj.weight)
        self.l_proj.bias.data.fill_(0)
        nn.init.xavier_uniform_(self.values_v_proj.weight)
        self.values_v_proj.bias.data.fill_(0)
        nn.init.xavier_uniform_(self.values_l_proj.weight)
        self.values_l_proj.bias.data.fill_(0)
        nn.init.xavier_uniform_(self.out_v_proj.weight)
        self.out_v_proj.bias.data.fill_(0)
        nn.init.xavier_uniform_(self.out_l_proj.weight)
        self.out_l_proj.bias.data.fill_(0)

    def forward(self, v, l, attention_mask_v=None, attention_mask_l=None):
        """_summary_

        Args:
            v (_type_): bs, n_img, dim
            l (_type_): bs, n_text, dim
            attention_mask_v (_type_, optional): _description_. bs, n_img
            attention_mask_l (_type_, optional): _description_. bs, n_text

        Returns:
            _type_: _description_
        """
        # if os.environ.get('IPDB_SHILONG_DEBUG', None) == 'INFO':
        #     import ipdb; ipdb.set_trace()
        bsz, tgt_len, _ = v.size()

        query_states = self.v_proj(v) * self.scale
        key_states = self._shape(self.l_proj(l), -1, bsz)
        value_v_states = self._shape(self.values_v_proj(v), -1, bsz)
        value_l_states = self._shape(self.values_l_proj(l), -1, bsz)

        proj_shape = (bsz * self.num_heads, -1, self.head_dim)
        query_states = self._shape(query_states, tgt_len, bsz).view(*proj_shape)
        key_states = key_states.view(*proj_shape)
        value_v_states = value_v_states.view(*proj_shape)
        value_l_states = value_l_states.view(*proj_shape)

        src_len = key_states.size(1)
        attn_weights = torch.bmm(
            query_states, key_states.transpose(1, 2)
        )  # bs*nhead, nimg, ntxt

        if attn_weights.size() != (bsz * self.num_heads, tgt_len, src_len):
            raise ValueError(
                f"Attention weights should be of size {(bsz * self.num_heads, tgt_len, src_len)}, but is {attn_weights.size()}"
            )

        if self.stable_softmax_2d:
            attn_weights = attn_weights - attn_weights.max()

        if self.clamp_min_for_underflow:
            attn_weights = torch.clamp(
                attn_weights, min=-50000
            )  # Do not increase -50000, data type half has quite limited range
        if self.clamp_max_for_overflow:
            attn_weights = torch.clamp(
                attn_weights, max=50000
            )  # Do not increase 50000, data type half has quite limited range

        attn_weights_T = attn_weights.transpose(1, 2)
        attn_weights_l = (
            attn_weights_T - torch.max(attn_weights_T, dim=-1, keepdim=True)[0]
        )
        if self.clamp_min_for_underflow:
            attn_weights_l = torch.clamp(
                attn_weights_l, min=-50000
            )  # Do not increase -50000, data type half has quite limited range
        if self.clamp_max_for_overflow:
            attn_weights_l = torch.clamp(
                attn_weights_l, max=50000
            )  # Do not increase 50000, data type half has quite limited range

        # mask vison for language
        if attention_mask_v is not None:
            attention_mask_v = (
                attention_mask_v[:, None, None, :]
                .repeat(1, self.num_heads, 1, 1)
                .flatten(0, 1)
            )
            attn_weights_l.masked_fill_(attention_mask_v, float("-inf"))

        attn_weights_l = attn_weights_l.softmax(dim=-1)

        # mask language for vision
        if attention_mask_l is not None:
            attention_mask_l = (
                attention_mask_l[:, None, None, :]
                .repeat(1, self.num_heads, 1, 1)
                .flatten(0, 1)
            )
            attn_weights.masked_fill_(attention_mask_l, float("-inf"))
        attn_weights_v = attn_weights.softmax(dim=-1)

        attn_probs_v = F.dropout(attn_weights_v, p=self.dropout, training=self.training)
        attn_probs_l = F.dropout(attn_weights_l, p=self.dropout, training=self.training)

        attn_output_v = torch.bmm(attn_probs_v, value_l_states)
        attn_output_l = torch.bmm(attn_probs_l, value_v_states)

        if attn_output_v.size() != (bsz * self.num_heads, tgt_len, self.head_dim):
            raise ValueError(
                f"`attn_output_v` should be of size {(bsz, self.num_heads, tgt_len, self.head_dim)}, but is {attn_output_v.size()}"
            )

        if attn_output_l.size() != (bsz * self.num_heads, src_len, self.head_dim):
            raise ValueError(
                f"`attn_output_l` should be of size {(bsz, self.num_heads, src_len, self.head_dim)}, but is {attn_output_l.size()}"
            )

        attn_output_v = attn_output_v.view(bsz, self.num_heads, tgt_len, self.head_dim)
        attn_output_v = attn_output_v.transpose(1, 2)
        attn_output_v = attn_output_v.reshape(bsz, tgt_len, self.embed_dim)

        attn_output_l = attn_output_l.view(bsz, self.num_heads, src_len, self.head_dim)
        attn_output_l = attn_output_l.transpose(1, 2)
        attn_output_l = attn_output_l.reshape(bsz, src_len, self.embed_dim)

        attn_output_v = self.out_v_proj(attn_output_v)
        attn_output_l = self.out_l_proj(attn_output_l)

        return attn_output_v, attn_output_l


# Bi-Direction MHA (text->image, image->text)
class BiAttentionBlock(nn.Module):
    def __init__(
        self,
        v_dim,
        l_dim,
        embed_dim,
        num_heads,
        dropout=0.1,
        drop_path=0.0,
        init_values=1e-4,
        cfg=None,
    ):
        """
        Inputs:
            embed_dim - Dimensionality of input and attention feature vectors
            hidden_dim - Dimensionality of hidden layer in feed-forward network
                         (usually 2-4x larger than embed_dim)
            num_heads - Number of heads to use in the Multi-Head Attention block
            dropout - Amount of dropout to apply in the feed-forward network
        """
        super(BiAttentionBlock, self).__init__()

        # pre layer norm
        self.layer_norm_v = nn.LayerNorm(v_dim)
        self.layer_norm_l = nn.LayerNorm(l_dim)
        self.attn = BiMultiHeadAttention(
            v_dim=v_dim,
            l_dim=l_dim,
            embed_dim=embed_dim,
            num_heads=num_heads,
            dropout=dropout,
        )

        # add layer scale for training stability
        self.drop_path = DropPath(drop_path) if drop_path > 0.0 else nn.Identity()
        self.gamma_v = nn.Parameter(
            init_values * torch.ones((v_dim)), requires_grad=True
        )
        self.gamma_l = nn.Parameter(
            init_values * torch.ones((l_dim)), requires_grad=True
        )

    def forward(self, v, l, attention_mask_v=None, attention_mask_l=None):
        v = self.layer_norm_v(v)
        l = self.layer_norm_l(l)
        delta_v, delta_l = self.attn(
            v, l, attention_mask_v=attention_mask_v, attention_mask_l=attention_mask_l
        )
        # v, l = v + delta_v, l + delta_l
        v = v + self.drop_path(self.gamma_v * delta_v)
        l = l + self.drop_path(self.gamma_l * delta_l)
        return v, l

    # def forward(self, v:List[torch.Tensor], l, attention_mask_v=None, attention_mask_l=None)



# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Temporal Shift Module w/ ResNet-50 and ResNet-101.

Based on:
  TSM: Temporal Shift Module for Efficient Video Understanding
  Ji Lin, Chuang Gan, Song Han
  https://arxiv.org/pdf/1811.08383.pdf.
"""

from typing import Optional, Sequence, Union
from absl import logging

import chex
import haiku as hk
import jax
import jax.numpy as jnp
import typing_extensions

from juxtapose.trackers.tapnet.models import tsm_utils as tsmu


class NormalizeFn(typing_extensions.Protocol):
    def __call__(self, x: chex.Array, is_training: bool) -> chex.Array:
        pass


class TSMResNetBlock(hk.Module):
    """A ResNet subblock with Temporal Channel Shifting.

    Combines a typical ResNetV2 block implementation
    (see https://arxiv.org/abs/1512.03385) with a pre-convolution Temporal
    Shift Module (see https://arxiv.org/pdf/1811.08383.pdf) in the residual.
    """

    def __init__(
        self,
        output_channels: int,
        stride: int,
        use_projection: bool,
        tsm_mode: str,
        normalize_fn: Optional[NormalizeFn] = None,
        channel_shift_fraction: float = 0.125,
        num_frames: int = 8,
        rate: int = 1,
        use_bottleneck: bool = False,
        name: str = "TSMResNetBlock",
    ):
        """Initializes the TSMResNetBlock module.

        Args:
          output_channels: Number of output channels.
          stride: Stride used in convolutions.
          use_projection: Whether to use a projection for the shortcut.
          tsm_mode: Mode for TSM ('gpu' or 'tpu' or 'deflated_0.x').
          normalize_fn: Function used for normalization.
          channel_shift_fraction: The fraction of temporally shifted channels. If
            `channel_shift_fraction` is 0, the block is the same as a normal ResNet
            block.
          num_frames: Size of frame dimension in a single batch example.
          rate: dilation rate.
          use_bottleneck: use a bottleneck (resnet-50 and above),
          name: The name of the module.
        """
        super().__init__(name=name)
        self._output_channels = (
            output_channels if use_bottleneck else output_channels // 4
        )
        self._bottleneck_channels = output_channels // 4
        self._stride = stride
        self._rate = rate
        self._use_projection = use_projection
        self._normalize_fn = normalize_fn
        self._tsm_mode = tsm_mode
        self._channel_shift_fraction = channel_shift_fraction
        self._num_frames = num_frames
        self._use_bottleneck = use_bottleneck

    def __call__(
        self,
        inputs: chex.Array,
        is_training: bool = True,
    ) -> chex.Array:
        """Connects the ResNetBlock module into the graph.

        Args:
          inputs: A 4-D float array of shape `[B, H, W, C]`.
          is_training: Whether to use training mode.

        Returns:
          A 4-D float array of shape
          `[B * num_frames, new_h, new_w, output_channels]`.
        """
        # ResNet V2 uses pre-activation, where the batch norm and relu are before
        # convolutions, rather than after as in ResNet V1.
        preact = inputs
        if self._normalize_fn is not None:
            preact = self._normalize_fn(preact, is_training=is_training)
        preact = jax.nn.relu(preact)

        if self._use_projection:
            shortcut = hk.Conv2D(
                output_channels=self._output_channels,
                kernel_shape=1,
                stride=self._stride,
                with_bias=False,
                padding="SAME",
                name="shortcut_conv",
            )(preact)
        else:
            shortcut = inputs

        # Eventually applies Temporal Shift Module.
        if self._channel_shift_fraction != 0:
            preact = tsmu.apply_temporal_shift(
                preact,
                tsm_mode=self._tsm_mode,
                num_frames=self._num_frames,
                channel_shift_fraction=self._channel_shift_fraction,
            )

        # First convolution.
        residual = hk.Conv2D(
            self._bottleneck_channels,
            kernel_shape=1 if self._use_bottleneck else 3,
            stride=1 if self._use_bottleneck else self._stride,
            with_bias=False,
            padding="SAME",
            name="conv_0",
        )(preact)

        if self._use_bottleneck:
            # Second convolution.
            if self._normalize_fn is not None:
                residual = self._normalize_fn(residual, is_training=is_training)
            residual = jax.nn.relu(residual)
            residual = hk.Conv2D(
                output_channels=self._bottleneck_channels,
                kernel_shape=3,
                stride=self._stride,
                rate=self._rate,
                with_bias=False,
                padding="SAME",
                name="conv_1",
            )(residual)

        # Third convolution.
        if self._normalize_fn is not None:
            residual = self._normalize_fn(residual, is_training=is_training)
        residual = jax.nn.relu(residual)
        residual = hk.Conv2D(
            output_channels=self._output_channels,
            kernel_shape=1 if self._use_bottleneck else 3,
            stride=1,
            with_bias=False,
            padding="SAME",
            name="conv_2",
        )(residual)

        # NOTE: we do not use block multiplier.
        output = shortcut + residual
        return output


class TSMResNetUnit(hk.Module):
    """Block group for TSM ResNet."""

    def __init__(
        self,
        output_channels: int,
        num_blocks: int,
        stride: int,
        tsm_mode: str,
        num_frames: int,
        normalize_fn: Optional[NormalizeFn] = None,
        channel_shift_fraction: float = 0.125,
        rate: int = 1,
        use_bottleneck: bool = False,
        name: str = "tsm_resnet_unit",
    ):
        """Creates a TSMResNet Unit.

        Args:
          output_channels: Number of output channels.
          num_blocks: Number of ResNet blocks in the unit.
          stride: Stride of the unit.
          tsm_mode: Which temporal shift module to use.
          num_frames: Size of frame dimension in a single batch example.
          normalize_fn: Function used for normalization.
          channel_shift_fraction: The fraction of temporally shifted channels. If
            `channel_shift_fraction` is 0, the block is the same as a normal ResNet
            block.
          rate: dilation rate.
          use_bottleneck: use a bottleneck (resnet-50 and above).
          name: The name of the module.
        """
        super().__init__(name=name)
        self._output_channels = output_channels
        self._num_blocks = num_blocks
        self._normalize_fn = normalize_fn
        self._stride = stride
        self._tsm_mode = tsm_mode
        self._channel_shift_fraction = channel_shift_fraction
        self._num_frames = num_frames
        self._rate = rate
        self._use_bottleneck = use_bottleneck

    def __call__(
        self,
        inputs: chex.Array,
        is_training: bool,
    ) -> chex.Array:
        """Connects the module to inputs.

        Args:
          inputs: A 4-D float array of shape `[B * num_frames, H, W, C]`.
          is_training: Whether to use training mode.

        Returns:
          A 4-D float array of shape
          `[B * num_frames, H // stride, W // stride, output_channels]`.
        """
        net = inputs
        for idx_block in range(self._num_blocks):
            net = TSMResNetBlock(
                self._output_channels,
                stride=(self._stride if idx_block == 0 else 1),
                rate=(max(self._rate // 2, 1) if idx_block == 0 else self._rate),
                use_projection=(idx_block == 0),
                normalize_fn=self._normalize_fn,
                tsm_mode=self._tsm_mode,
                channel_shift_fraction=self._channel_shift_fraction,
                num_frames=self._num_frames,
                name=f"block_{idx_block}",
            )(net, is_training=is_training)
        return net


class TSMResNetV2(hk.Module):
    """TSM based on ResNet V2 as described in https://arxiv.org/abs/1603.05027."""

    # Endpoints of the model in order.
    VALID_ENDPOINTS = (
        "tsm_resnet_stem",
        "tsm_resnet_unit_0",
        "tsm_resnet_unit_1",
        "tsm_resnet_unit_2",
        "tsm_resnet_unit_3",
        "last_conv",
        "Embeddings",
    )

    def __init__(
        self,
        normalize_fn: Optional[NormalizeFn] = None,
        depth: int = 18,
        num_frames: int = 16,
        channel_shift_fraction: Union[float, Sequence[float]] = 0.125,
        width_mult: int = 1,
        name: str = "TSMResNetV2",
    ):
        """Constructs a ResNet model.

        Args:
          normalize_fn: Function used for normalization.
          depth: Depth of the desired ResNet.
          num_frames: Number of frames (used in TPU mode).
          channel_shift_fraction: Fraction of channels that are temporally shifted,
            if `channel_shift_fraction` is 0, a regular ResNet is returned.
          width_mult: Whether or not to use a width multiplier.
          name: The name of the module.

        Raises:
          ValueError: If `channel_shift_fraction` or `depth` has invalid value.
        """
        super().__init__(name=name)

        if isinstance(channel_shift_fraction, float):
            channel_shift_fraction = [channel_shift_fraction] * 4

        if not all([0.0 <= x <= 1.0 for x in channel_shift_fraction]):
            raise ValueError(
                f"channel_shift_fraction ({channel_shift_fraction})"
                " all have to be in [0, 1]."
            )

        self._num_frames = num_frames

        self._channels = (256, 512, 1024, 2048)

        num_blocks = {
            18: (2, 2, 2, 2),
            34: (3, 4, 6, 3),
            50: (3, 4, 6, 3),
            101: (3, 4, 23, 3),
            152: (3, 8, 36, 3),
            200: (3, 24, 36, 3),
        }
        if depth not in num_blocks:
            raise ValueError(
                f"`depth` should be in {list(num_blocks.keys())} ({depth} given)."
            )
        self._num_blocks = num_blocks[depth]

        self._width_mult = width_mult
        self._channel_shift_fraction = channel_shift_fraction
        self._normalize_fn = normalize_fn
        self._use_bottleneck = depth >= 50

    def __call__(
        self,
        inputs: chex.Array,
        is_training: bool = True,
        final_endpoint: str = "Embeddings",
        is_deflated: bool = False,
        alpha_deflation: float = 0.3,
        out_num_frames: Optional[int] = None,
        output_stride: int = 8,
    ) -> chex.Array:
        """Connects the TSM ResNetV2 module into the graph.

        Args:
          inputs: The input may be in one of two shapes; if the shape is `[B, T, H,
            W, C]`, this module assumes the backend is a GPU (setting
            `tsm_mode='gpu'`) and `T` is treated the time dimension, with `B` being
            the batch dimension. This mode cannot be used when `is_deflated` is
            `true`. In this mode, the num_frames parameter passed to the constructor
            is ignored. If the shape is `[B, H, W, C]`, then the batch dimension is
            assumed to be of the form [B*T, H, W, C], where `T` is the number of
            frames in each video. This value may be set by passing `num_frames=n` to
            the constructor. The default value is `n=16` (beware this default is not
            the same as the default for the `TSMResNetBlock`, which has a default of
            8 frames). In this case, the module assumes it is being run on a TPU,
            and emits instructions that are more efficient for that case,
            using`tsm_mode`='tpu'` for the downstream blocks.
          is_training: Whether to use training mode.
          final_endpoint: Up to which endpoint to run / return.
          is_deflated: Whether or not to use the deflated version of the network.
          alpha_deflation: Deflation parameter to use for dealing with the padding
            effect.
          out_num_frames: Whether time is on first axis, for TPU performance
          output_stride: Stride of the final feature grid; possible values are
            4, 8, 16, or 32.  32 is the standard for TSM-ResNet. Others strides are
            achieved by converting strided to un-strided convolutions later in the
            network, while increasing the dilation rate for later layers.

        Returns:
          Network output at location `final_endpoint`. A float array which shape
          depends on `final_endpoint`.

        Raises:
          ValueError: If `final_endpoint` is not recognized.
        """

        # Prepare inputs for TSM.
        if is_deflated:
            if len(inputs.shape) != 4:
                raise ValueError(
                    "In deflated mode inputs should be given as [B, H, W, 3]"
                )
            logging.warning(
                "Deflation is an experimental feature and the API might change."
            )
            tsm_mode = f"deflated_{alpha_deflation}"
            num_frames = 1
        else:
            inputs, tsm_mode, num_frames = tsmu.prepare_inputs(inputs)
            num_frames = num_frames or out_num_frames or self._num_frames

        self._final_endpoint = final_endpoint
        if self._final_endpoint not in self.VALID_ENDPOINTS:
            raise ValueError(f"Unknown final endpoint {self._final_endpoint}")

        # Stem convolution.
        end_point = "tsm_resnet_stem"
        net = hk.Conv2D(
            output_channels=64 * self._width_mult,
            kernel_shape=7,
            stride=2,
            with_bias=False,
            name=end_point,
            padding="SAME",
        )(inputs)
        net = hk.MaxPool(
            window_shape=(1, 3, 3, 1),
            strides=(1, 2, 2, 1),
            padding="SAME",
        )(net)
        if self._final_endpoint == end_point:
            net = tsmu.prepare_outputs(net, tsm_mode, num_frames, reduce_mean=False)
            return net

        if output_stride == 4:
            strides = (1, 1, 1, 1)
            rates = (1, 2, 4, 8)
        elif output_stride == 8:
            strides = (1, 2, 1, 1)
            rates = (1, 1, 2, 4)
        elif output_stride == 16:
            strides = (1, 2, 2, 1)
            rates = (1, 1, 1, 2)
        elif output_stride == 32:
            strides = (1, 2, 2, 2)
            rates = (1, 1, 1, 1)
        else:
            raise ValueError("unsupported output stride")

        # Residual block.
        for unit_id, (channels, num_blocks, stride, rate) in enumerate(
            zip(self._channels, self._num_blocks, strides, rates)
        ):
            end_point = f"tsm_resnet_unit_{unit_id}"
            net = TSMResNetUnit(
                output_channels=channels * self._width_mult,
                num_blocks=num_blocks,
                stride=stride,
                rate=rate,
                normalize_fn=self._normalize_fn,
                channel_shift_fraction=self._channel_shift_fraction[unit_id],
                num_frames=num_frames,
                tsm_mode=tsm_mode,
                use_bottleneck=self._use_bottleneck,
                name=end_point,
            )(net, is_training=is_training)
            if self._final_endpoint == end_point:
                net = tsmu.prepare_outputs(net, tsm_mode, num_frames, reduce_mean=False)
                return net

        if self._normalize_fn is not None:
            net = self._normalize_fn(net, is_training=is_training)
        net = jax.nn.relu(net)

        end_point = "last_conv"
        if self._final_endpoint == end_point:
            net = tsmu.prepare_outputs(net, tsm_mode, num_frames, reduce_mean=False)
            return net
        net = jnp.mean(net, axis=(1, 2))
        # Prepare embedding outputs for TSM (temporal average of features).
        net = tsmu.prepare_outputs(net, tsm_mode, num_frames, reduce_mean=True)
        assert self._final_endpoint == "Embeddings"
        return net



from ultralytics import YOLO
import numpy as np

from juxtapose.utils import LOGGER
from juxtapose.utils.core import Detections

import torch
from juxtapose.utils.downloads import safe_download

base = "juxtapose"


class YOLOv8:
    """YOLOv8 model (s, m, l) to detect only person (class 0)"""

    def __init__(self, type: str = "m", device: str = "cpu", conf_thres: float = 0.3):
        self.model = YOLO(f"model/yolov8{type}.pt")
        if device == "cuda" and torch.cuda.is_available():
            self.model.to(device)
        else:
            self.model.to("cpu")
        self.conf_thres = conf_thres
        LOGGER.info(f"Loaded yolov8{type} pt model.")

    def __call__(self, im):
        """Return List of xyxy coordinates based on im frame (cv2.imread)
        im -> (h, w, c)
        return -> [[x, y, x, y], [x, y, x, y], ...] -> two persons detected
        """
        result = self.model(im, verbose=False, conf=self.conf_thres)[0].boxes
        result = result[result.cls == 0].cpu().numpy()
        result = Detections(
            xyxy=result.xyxy,
            confidence=result.conf,
            labels=np.array([0 for _ in range(len(result.xyxy))]),
        )
        # print(result)

        return result



import cv2
import torch
import bisect
import numpy as np
from pathlib import Path
from juxtapose.utils.core import Detections
from juxtapose.utils import ops


from juxtapose.utils.downloads import safe_download
import juxtapose.detectors.groundingdino.datasets.transforms as T
from juxtapose.detectors.groundingdino.util.inference import (
    load_model,
    preprocess_caption,
    get_phrases_from_posmap,
    box_convert,
)


class GroundingDino:
    def __init__(
        self,
        type="",
        device: str = "cpu",
        captions: str = "person .",
        conf_thres: float = 0.35,
        text_thres: float = 0.25,
    ):
        model_path = Path(f"model/groundingdino_swint_ogc.pth")
        config_path = Path(f"model/gd-ogc.py")
        if not model_path.exists():
            safe_download(
                f"https://huggingface.co/ziq/rtm/resolve/main/groundingdino_swint_ogc.pth",
                file=f"groundingdino_swint_ogc.pth",
                dir=Path(f"model/"),
            )
        if not config_path.exists():
            safe_download(
                f"https://huggingface.co/ziq/rtm/resolve/main/gd-ogc.py",
                file=f"gd-ogc.py",
                dir=Path(f"model/"),
            )
        self.model = load_model(str(config_path), model_path, captions=captions)
        self.model.to(device)
        self.device = device
        self.captions = captions
        self.conf_thres = conf_thres
        self.text_thres = text_thres

    def transform(self, im):
        _transform = T.Compose(
            [
                T.ToPILImage(),
                T.RandomResize([800], max_size=1333),
                T.ToTensor(),
                T.Normalize([0.485, 0.456, 0.406], [0.229, 0.224, 0.225]),
            ]
        )
        im, _ = _transform(im, None)
        return im

    def predict(self, im, remove_combined: bool = False):
        with torch.no_grad():
            outputs = self.model(im[None])
            # outputs = self.model(im[None], captions=[self.captions])

        prediction_logits = (
            outputs["pred_logits"].cpu().sigmoid()[0]
        )  # prediction_logits.shape = (nq, 256)
        prediction_boxes = outputs["pred_boxes"].cpu()[
            0
        ]  # prediction_boxes.shape = (nq, 4)

        mask = prediction_logits.max(dim=1)[0] > self.conf_thres
        logits = prediction_logits[mask]  # logits.shape = (n, 256)
        boxes = prediction_boxes[mask]  # boxes.shape = (n, 4)

        tokenizer = self.model.tokenizer
        tokenized = tokenizer(self.captions)

        if remove_combined:
            sep_idx = [
                i
                for i in range(len(tokenized["input_ids"]))
                if tokenized["input_ids"][i] in [101, 102, 1012]
            ]

            phrases = []
            for logit in logits:
                max_idx = logit.argmax()
                insert_idx = bisect.bisect_left(sep_idx, max_idx)
                right_idx = sep_idx[insert_idx]
                left_idx = sep_idx[insert_idx - 1]
                phrases.append(
                    get_phrases_from_posmap(
                        logit > self.text_thres,
                        tokenized,
                        tokenizer,
                        left_idx,
                        right_idx,
                    ).replace(".", "")
                )
        else:
            phrases = [
                get_phrases_from_posmap(
                    logit > self.text_thres, tokenized, tokenizer
                ).replace(".", "")
                for logit in logits
            ]

        return boxes, logits.max(dim=1)[0], phrases

    def __call__(self, im):
        h, w, _ = im.shape
        im = cv2.cvtColor(im, cv2.COLOR_BGR2RGB)
        im = self.transform(im)
        bboxes, conf, labels = self.predict(im)
        bboxes = bboxes * torch.Tensor([w, h, w, h])
        xyxy = box_convert(boxes=bboxes, in_fmt="cxcywh", out_fmt="xyxy").numpy()

        result = Detections(
            xyxy=xyxy,
            confidence=conf.numpy(),
            labels=np.array([0 if label == "person" else 0 for label in labels]),
            # labels=np.array([0 if label == "person" else -1 for label in labels]),
        )
        # print(result)
        return result



import os
import random
from pathlib import Path

import numpy as np
import torch
from PIL import Image

from juxtapose.data.loaders import (
    LOADERS,
    LoadImages,
    LoadPilAndNumpy,
    LoadScreenshots,
    LoadStreams,
    LoadTensor,
    SourceTypes,
    autocast_list,
)

from juxtapose.data.utils import IMG_FORMATS, VID_FORMATS
from juxtapose.utils import RANK, colorstr
from juxtapose.utils.checks import check_file


def check_source(source):
    """Check source type and return corresponding flag values."""
    webcam, screenshot, from_img, in_memory, tensor = False, False, False, False, False
    if isinstance(source, (str, int, Path)):  # int for local usb camera
        source = str(source)
        is_file = Path(source).suffix[1:] in (IMG_FORMATS + VID_FORMATS)
        is_url = source.lower().startswith(
            ("https://", "http://", "rtsp://", "rtmp://")
        )
        webcam = (
            source.isnumeric()
            or source.endswith(".streams")
            or (is_url and not is_file)
        )
        screenshot = source.lower() == "screen"
        if is_url and is_file:
            source = check_file(source)  # download
    elif isinstance(source, tuple(LOADERS)):
        in_memory = True
    elif isinstance(source, (list, tuple)):
        source = autocast_list(source)  # convert all list elements to PIL or np arrays
        from_img = True
    elif isinstance(source, (Image.Image, np.ndarray)):
        from_img = True
    elif isinstance(source, torch.Tensor):
        tensor = True
    else:
        raise TypeError(
            "Unsupported image type. For supported types see https://docs.ultralytics.com/modes/predict"
        )

    return source, webcam, screenshot, from_img, in_memory, tensor


def load_inference_source(source=None, imgsz=640, vid_stride=1):
    """
    Loads an inference source for object detection and applies necessary transformations.

    Args:
        source (str, Path, Tensor, PIL.Image, np.ndarray): The input source for inference.
        imgsz (int, optional): The size of the image for inference. Default is 640.
        vid_stride (int, optional): The frame interval for video sources. Default is 1.

    Returns:
        dataset (Dataset): A dataset object for the specified input source.
    """
    source, webcam, screenshot, from_img, in_memory, tensor = check_source(source)
    source_type = (
        source.source_type
        if in_memory
        else SourceTypes(webcam, screenshot, from_img, tensor)
    )

    # Dataloader
    if tensor:
        dataset = LoadTensor(source)
    elif in_memory:
        dataset = source
    elif webcam:
        dataset = LoadStreams(source, imgsz=imgsz, vid_stride=vid_stride)
    elif screenshot:
        dataset = LoadScreenshots(source, imgsz=imgsz)
    elif from_img:
        dataset = LoadPilAndNumpy(source, imgsz=imgsz)
    else:
        dataset = LoadImages(source, imgsz=imgsz, vid_stride=vid_stride)

    # Attach source types to the dataset
    setattr(dataset, "source_type", source_type)

    return dataset



__version__ = "0.0.32"

from .rtm import RTM
from .detectors import RTMDet, YOLOv8

from .pose.rtmpose import RTMPose
from .tapnet import Tapnet
from .utils.plotting import Annotator

__all__ = (
    "__version__",
    "RTM",
    "RTMDet",
    "RTMPose",
    "Tapnet",
    "Detections",
    "Annotator",
    "GroundingDino",
    "YOLOv8",
)



# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""TAPIR model definition."""

import functools
from typing import Any, Dict, List, Mapping, NamedTuple, Optional, Sequence, Tuple

import chex
from einshape import jax_einshape as einshape
import haiku as hk
import jax
import jax.numpy as jnp

from juxtapose.trackers.tapnet.models import resnet
from juxtapose.trackers.tapnet.utils import model_utils
from juxtapose.trackers.tapnet.utils import transforms


def layernorm(x):
    return hk.LayerNorm(axis=-1, create_scale=True, create_offset=False)(x)


def depthwise_conv_residual(
    x: chex.Array,
    kernel_shape: int = 3,
    use_causal_conv: bool = False,
    causal_context: Optional[Mapping[str, chex.Array]] = None,
    get_causal_context: bool = False,
) -> Tuple[chex.Array, Dict[str, chex.Array]]:
    """First mlp in mixer."""
    cur_name = hk.experimental.current_name()
    name1 = cur_name + "_causal_1"
    if causal_context is not None:
        x = jnp.concatenate([causal_context[name1], x], axis=-2)
        # Because we're adding extra frames, the output of this convolution will
        # also have extra (invalid) frames that need to be removed afterward.
        num_extra = causal_context[name1].shape[-2]
    new_causal_context = {}
    if get_causal_context:
        # Keep only as many frames of x as needed for future frames.  This may be up
        # to (kernel_shape - 1) frames.
        new_causal_context[name1] = x[..., -(kernel_shape - 1) :, :]
    x = hk.DepthwiseConv1D(
        channel_multiplier=4,
        kernel_shape=kernel_shape,
        data_format="NWC",
        stride=1,
        padding=[(kernel_shape - 1, 0)] if use_causal_conv else "SAME",
        name="mlp1_up",
    )(x)
    x = jax.nn.gelu(x)
    name2 = cur_name + "_causal_2"
    if causal_context is not None:
        x = jnp.concatenate([causal_context[name2], x[..., num_extra:, :]], axis=-2)
        num_extra = causal_context[name2].shape[-2]
    if get_causal_context:
        new_causal_context[name2] = x[..., -(kernel_shape - 1) :, :]

    x = hk.DepthwiseConv1D(
        channel_multiplier=1,
        kernel_shape=kernel_shape,
        data_format="NWC",
        stride=1,
        padding=[(kernel_shape - 1, 0)] if use_causal_conv else "SAME",
        name="mlp1_up",
    )(x)
    if causal_context is not None:
        x = x[..., num_extra:, :]

    return (
        x[..., 0::4] + x[..., 1::4] + x[..., 2::4] + x[..., 3::4],
        new_causal_context,
    )


def conv_channels_mixer(x):
    """Second mlp in mixer."""
    in_channels = x.shape[-1]
    x = hk.Linear(in_channels * 4, name="mlp2_up")(x)
    x = jax.nn.gelu(x)
    x = hk.Linear(in_channels, name="mlp2_down")(x)
    return x


class PIPsConvBlock(hk.Module):
    """Transformer block (mha and ffw)."""

    def __init__(self, name="block", kernel_shape=3, use_causal_conv=False):
        super().__init__(name=name)
        self.kernel_shape = kernel_shape
        self.use_causal_conv = use_causal_conv

    def __call__(self, x, causal_context=None, get_causal_context=False):
        to_skip = x
        x = layernorm(x)
        x, new_causal_context = depthwise_conv_residual(
            x,
            self.kernel_shape,
            self.use_causal_conv,
            causal_context,
            get_causal_context,
        )
        x = x + to_skip
        to_skip = x
        x = layernorm(x)
        x = conv_channels_mixer(x)
        x = x + to_skip
        return x, new_causal_context


class PIPSMLPMixer(hk.Module):
    """Depthwise-conv version of PIPs's MLP Mixer."""

    def __init__(
        self,
        output_channels,
        hidden_dim=512,
        num_blocks=12,
        kernel_shape=3,
        use_causal_conv=False,
        name="pips_mlp_mixer",
    ):
        super().__init__(name=name)
        self._output_channels = output_channels
        self.hidden_dim = hidden_dim
        self._num_blocks = num_blocks
        self.kernel_shape = kernel_shape
        self.use_causal_conv = use_causal_conv

    def __call__(self, x, causal_context=None, get_causal_context=False):
        x = hk.Linear(self.hidden_dim)(x)
        all_causal_context = {}
        for _ in range(self._num_blocks):
            x, new_causal_context = PIPsConvBlock(
                kernel_shape=self.kernel_shape, use_causal_conv=self.use_causal_conv
            )(x, causal_context, get_causal_context)
            if get_causal_context:
                all_causal_context.update(new_causal_context)
        x = layernorm(x)
        return hk.Linear(self._output_channels)(x), all_causal_context


def construct_patch_kernel(pos, grid_size, patch_size=7):
    """A conv kernel that performs bilinear interpolation for a point."""
    # pos is n-by-2, [y,x]
    # grid_size is [heigh,width]
    # result is [1,n,kernel_height,kernel_width]
    pos = pos + (patch_size) / 2 - 1

    def gen_bump(pos, num):
        # pos is shape [n]
        # result is shape [n,num]
        res = jnp.arange(num)
        return jnp.maximum(0, 1 - jnp.abs(res[jnp.newaxis, :] - pos[:, jnp.newaxis]))

    x_bump = gen_bump(pos[:, 1], grid_size[1] - patch_size + 1)
    y_bump = gen_bump(pos[:, 0], grid_size[0] - patch_size + 1)

    kernel = (
        x_bump[:, jnp.newaxis, jnp.newaxis, :] * y_bump[:, jnp.newaxis, :, jnp.newaxis]
    )
    return kernel


def extract_patch_depthwise_conv(pos, corrs, patch_size=7):
    """Use a depthwise conv to extract a patch via bilinear interpolation."""
    # pos is n-by-2, [y,x], raster coordinates
    # arr is [num_points, height, width]
    # result is [num_points, height, width]
    # add an extra batch axis because conv needs it
    corrs = jnp.pad(
        corrs,
        (
            (0, 0),
            (patch_size - 1, patch_size - 1),
            (patch_size - 1, patch_size - 1),
        ),
    )[jnp.newaxis, ...]
    kernel = construct_patch_kernel(pos, corrs.shape[2:4], patch_size)
    dim_nums = jax.lax.ConvDimensionNumbers(
        lhs_spec=(0, 1, 2, 3), rhs_spec=(0, 1, 2, 3), out_spec=(0, 1, 2, 3)
    )
    # the [0] gets rid of the extra batch axis
    res = jax.lax.conv_general_dilated(
        corrs,
        kernel,
        (1, 1),
        "VALID",
        (1, 1),
        (1, 1),
        dim_nums,
        feature_group_count=kernel.shape[0],
    )[0]
    return res


def is_same_res(r1, r2):
    """Test if two image resolutions are the same."""
    return all([x == y for x, y in zip(r1, r2)])


class FeatureGrids(NamedTuple):
    """Feature grids for a video, used to compute trajectories.

    These are per-frame outputs of the encoding resnet.

    Attributes:
      lowres: Low-resolution features, one for each resolution; 256 channels.
      hires: High-resolution features, one for each resolution; 64 channels.
      resolutions: Resolutions used for trajectory computation.  There will be one
        entry for the initialization, and then an entry for each PIPs refinement
        resolution.
    """

    lowres: Sequence[chex.Array]
    hires: Sequence[chex.Array]
    resolutions: Sequence[Tuple[int, int]]


class QueryFeatures(NamedTuple):
    """Query features used to compute trajectories.

    These are sampled from the query frames and are a full descriptor of the
    tracked points. They can be acquired from a query image and then reused in a
    separate video.

    Attributes:
      lowres: Low-resolution features, one for each resolution; each has shape
        [batch, num_query_points, 256]
      hires: High-resolution features, one for each resolution; each has shape
        [batch, num_query_points, 64]
      resolutions: Resolutions used for trajectory computation.  There will be one
        entry for the initialization, and then an entry for each PIPs refinement
        resolution.
    """

    lowres: Sequence[chex.Array]
    hires: Sequence[chex.Array]
    resolutions: Sequence[Tuple[int, int]]


class TAPIR(hk.Module):
    """TAPIR model."""

    def __init__(
        self,
        bilinear_interp_with_depthwise_conv: bool = False,
        num_pips_iter: int = 4,
        pyramid_level: int = 1,
        mixer_hidden_dim: int = 512,
        num_mixer_blocks: int = 12,
        mixer_kernel_shape: int = 3,
        patch_size: int = 7,
        softmax_temperature: float = 20.0,
        use_causal_conv: bool = False,
        parallelize_query_extraction: bool = False,
        initial_resolution: Tuple[int, int] = (256, 256),
        name: str = "tapir",
    ):
        super().__init__(name=name)

        self.highres_dim = 128
        self.lowres_dim = 256

        self.resnet = resnet.ResNet(
            resnet_v2=True,
            normalization="instancenorm",
            strides=(1, 2, 2, 1),
            blocks_per_group=(2, 2, 2, 2),
            channels_per_group=(64, self.highres_dim, 256, self.lowres_dim),
            use_projection=(True, True, True, True),
            use_max_pool=False,
            name="resnet",
        )

        self.cost_volume_track_mods = {
            "hid1": hk.Conv2D(
                16,
                [3, 3],
                name="cost_volume_regression_1",
                stride=[1, 1],
            ),
            "hid2": hk.Conv2D(
                1,
                [3, 3],
                name="cost_volume_regression_2",
                stride=[1, 1],
            ),
            "hid3": hk.Conv2D(
                32,
                [3, 3],
                name="cost_volume_occlusion_1",
                stride=[2, 2],
            ),
            "hid4": hk.Linear(16, name="cost_volume_occlusion_2"),
            "occ_out": hk.Linear(2, name="occlusion_out"),
            "regression_hid": hk.Linear(128, name="regression_hid"),
            "regression_out": hk.Linear(2, name="regression_out"),
            "conv_stats_conv1": hk.Conv2D(
                256,
                [5, 5],
                name="conv_stats_conv1",
                stride=[1, 1],
            ),
            "conv_stats_conv2": hk.Conv2D(
                256,
                [3, 3],
                name="conv_stats_conv2",
                stride=[1, 1],
            ),
            "conv_stats_linear": hk.Linear(32, name="conv_stats_linear"),
        }
        self.pips_mixer = PIPSMLPMixer(
            4 + self.highres_dim + self.lowres_dim,
            hidden_dim=mixer_hidden_dim,
            num_blocks=num_mixer_blocks,
            kernel_shape=mixer_kernel_shape,
            use_causal_conv=use_causal_conv,
        )

        self.bilinear_interp_with_depthwise_conv = bilinear_interp_with_depthwise_conv
        self.parallelize_query_extraction = parallelize_query_extraction

        self.num_pips_iter = num_pips_iter
        self.pyramid_level = pyramid_level
        self.patch_size = patch_size
        self.softmax_temperature = softmax_temperature
        self.initial_resolution = tuple(initial_resolution)

    def tracks_from_cost_volume(
        self,
        interp_feature: chex.Array,
        feature_grid: chex.Array,
        query_points: Optional[chex.Array],
        im_shp: Optional[chex.Shape] = None,
    ) -> Tuple[chex.Array, chex.Array, chex.Array]:
        """Converts features into tracks by computing a cost volume.

        The computed cost volume will have shape
          [batch, num_queries, time, height, width], which can be very
          memory intensive.

        Args:
          interp_feature: A tensor of features for each query point, of shape
            [batch, num_queries, channels, heads].
          feature_grid: A tensor of features for the video, of shape [batch, time,
            height, width, channels, heads].
          query_points: When computing tracks, we assume these points are given as
            ground truth and we reproduce them exactly.  This is a set of points of
            shape [batch, num_points, 3], where each entry is [t, y, x] in frame/
            raster coordinates.
          im_shp: The shape of the original image, i.e., [batch, num_frames, time,
            height, width, 3].

        Returns:
          A 2-tuple of the inferred points (of shape
            [batch, num_points, num_frames, 2] where each point is [x, y]) and
            inferred occlusion (of shape [batch, num_points, num_frames], where
            each is a logit where higher means occluded)
        """

        mods = self.cost_volume_track_mods
        # Note: time is first axis to prevent the TPU from padding
        cost_volume = jnp.einsum(
            "bnc,bthwc->tbnhw",
            interp_feature,
            feature_grid,
        )
        shape = cost_volume.shape
        batch_size, num_points = cost_volume.shape[1:3]

        cost_volume = einshape("tbnhw->(tbn)hw1", cost_volume)

        occlusion = mods["hid1"](cost_volume)
        occlusion = jax.nn.relu(occlusion)

        pos = mods["hid2"](occlusion)
        pos_rshp = einshape("(tb)hw1->t(b)hw1", pos, t=shape[0])

        pos = einshape("t(bn)hw1->bnthw", pos_rshp, b=batch_size, n=num_points)
        pos = jax.nn.softmax(pos * self.softmax_temperature, axis=(-2, -1))
        points = model_utils.heatmaps_to_points(pos, im_shp, query_points=query_points)

        occlusion = mods["hid3"](occlusion)
        occlusion = jax.nn.relu(occlusion)
        occlusion = jnp.mean(occlusion, axis=(-2, -3))
        occlusion = mods["hid4"](occlusion)
        occlusion = jax.nn.relu(occlusion)
        occlusion = mods["occ_out"](occlusion)
        expected_dist = einshape(
            "(tbn)1->bnt", occlusion[..., 1:2], n=shape[2], t=shape[0]
        )
        occlusion = einshape("(tbn)1->bnt", occlusion[..., 0:1], n=shape[2], t=shape[0])
        return points, occlusion, expected_dist

    def refine_pips(
        self,
        target_feature,
        frame_features,
        pyramid,
        pos_guess,
        occ_guess,
        expd_guess,
        orig_hw,
        last_iter=None,
        mixer_iter=0.0,
        resize_hw=None,
        causal_context=None,
        get_causal_context=False,
    ):
        # frame_features is batch, num_frames, height, width, channels
        # target_features is batch, num_points, channels
        # pos_guess is batch, num_points, num_frames,2
        orig_h, orig_w = orig_hw
        resized_h, resized_w = resize_hw

        corrs_pyr = []
        assert len(target_feature) == len(pyramid)
        for pyridx, (query, grid) in enumerate(zip(target_feature, pyramid)):
            # note: interp needs [y,x]
            coords = transforms.convert_grid_coordinates(
                pos_guess, (orig_w, orig_h), grid.shape[-2:-4:-1]
            )[..., ::-1]
            last_iter_query = None
            if last_iter is not None:
                if pyridx == 0:
                    last_iter_query = last_iter[..., : self.highres_dim]
                else:
                    last_iter_query = last_iter[..., self.highres_dim :]
            if not self.bilinear_interp_with_depthwise_conv:
                # on CPU, gathers are cheap and matmuls are expensive
                ctxx, ctxy = jnp.meshgrid(jnp.arange(-3, 4), jnp.arange(-3, 4))
                ctx = jnp.stack([ctxy, ctxx], axis=-1)
                ctx = jnp.reshape(ctx, [-1, 2])
                coords2 = (
                    coords[:, :, :, jnp.newaxis, :]
                    + ctx[jnp.newaxis, jnp.newaxis, jnp.newaxis, ...]
                )
                # grid is batch, frames, height, width, channels
                # coords is batch, num_points, frames, spatial, x/y
                # neighborhood = batch, num_points, frames, patch_height, patch_width,
                # channels
                neighborhood = jax.vmap(  # across batch
                    jax.vmap(  # across frames
                        jax.vmap(  # across patch context size
                            jax.vmap(  # across channels
                                functools.partial(model_utils.interp, mode="constant"),
                                in_axes=(-1, None),
                                out_axes=-1,
                            ),
                            in_axes=(None, -2),
                            out_axes=-2,
                        ),
                        in_axes=(0, 1),
                        out_axes=1,
                    )
                )(grid, coords2)
                # s is spatial context size
                if last_iter_query is None:
                    patches = jnp.einsum("bnfsc,bnc->bnfs", neighborhood, query)
                else:
                    patches = jnp.einsum(
                        "bnfsc,bnfc->bnfs", neighborhood, last_iter_query
                    )
            else:
                # on TPU, matmul is cheap and gather is expensive, so we rewrite
                # the interpolation with a depthwise conv.
                if last_iter_query is None:
                    corrs = jnp.einsum("bfhwc,bnc->bnfhw", grid, query)
                else:
                    corrs = jnp.einsum("bfhwc,bnfc->bnfhw", grid, last_iter_query)
                n = corrs.shape[1]
                # coords is bnfs2
                # patches is batch,n,frames,height,width
                # vmap across batch dimension (because we have different points across
                # the batch) and across the frame axis (this could potentially be rolled
                # into the depthwise conv)
                extract_patch_depthwise_conv_ = functools.partial(
                    extract_patch_depthwise_conv, patch_size=self.patch_size
                )
                patches = jax.vmap(extract_patch_depthwise_conv_)(
                    einshape("bnfc->b(nf)c", coords), einshape("bnfhw->b(nf)hw", corrs)
                )
                patches = einshape("b(nf)hw->bnf(hw)", patches, n=n)
            corrs_pyr.append(patches)
        corrs_pyr = jnp.concatenate(corrs_pyr, axis=-1)

        corrs_chunked = corrs_pyr
        pos_guess_input = pos_guess
        occ_guess_input = occ_guess[..., jnp.newaxis]
        expd_guess_input = expd_guess[..., jnp.newaxis]

        # mlp_input is batch, num_points, num_chunks, frames_per_chunk, channels
        if last_iter is None:
            both_feature = jnp.concatenate(
                [target_feature[0], target_feature[1]], axis=-1
            )
            mlp_input_features = jnp.tile(
                both_feature[:, :, jnp.newaxis, :],
                (1, 1) + corrs_chunked.shape[-2:-1] + (1,),
            )
        else:
            mlp_input_features = last_iter

        pos_guess_input = jnp.zeros_like(pos_guess_input)

        mlp_input = jnp.concatenate(
            [
                pos_guess_input,
                occ_guess_input,
                expd_guess_input,
                mlp_input_features,
                corrs_chunked,
            ],
            axis=-1,
        )
        x = einshape("bnfc->(bn)fc", mlp_input)
        if causal_context is not None:
            causal_context = jax.tree_map(
                lambda x: einshape("bn...->(bn)...", x), causal_context
            )
        res, new_causal_context = self.pips_mixer(x, causal_context, get_causal_context)

        res = einshape("(bn)fc->bnfc", res, b=mlp_input.shape[0])
        if get_causal_context:
            new_causal_context = jax.tree_map(
                lambda x: einshape("(bn)...->bn...", x, b=mlp_input.shape[0]),
                new_causal_context,
            )

        pos_update = transforms.convert_grid_coordinates(
            res[..., :2],
            (resized_w, resized_h),
            (orig_w, orig_h),
        )
        return (
            pos_update + pos_guess,
            res[..., 2] + occ_guess,
            res[..., 3] + expd_guess,
            res[..., 4:] + (mlp_input_features if last_iter is None else last_iter),
            new_causal_context,
        )

    def get_feature_grids(
        self,
        video: chex.Array,
        is_training: bool,
        refinement_resolutions: Optional[List[Tuple[int, int]]] = None,
    ) -> FeatureGrids:
        """Computes feature grids.

        Args:
          video: A 5-D tensor representing a batch of sequences of images.
          is_training: Whether we are training.
          refinement_resolutions: A list of (height, width) tuples.  Refinement will
            be repeated at each specified resolution, in order to achieve high
            accuracy on resolutions higher than what TAPIR was trained on. If None,
            reasonable refinement resolutions will be inferred from the input video
            size.

        Returns:
          A FeatureGrids object which contains the required features for every
            required resolution.  Note that there will be one more feature grid
            than there are refinement_resolutions, because there is always a
            feature grid computed for TAP-Net initialization.
        """

        if refinement_resolutions is None:
            refinement_resolutions = model_utils.generate_default_resolutions(
                video.shape[2:4], self.initial_resolution
            )

        all_required_resolutions = [self.initial_resolution]
        all_required_resolutions.extend(refinement_resolutions)

        feature_grid = []
        hires_feats = []
        resize_im_shape = []
        curr_resolution = (-1, -1)
        for resolution in all_required_resolutions:
            if resolution[0] % 8 != 0 or resolution[1] % 8 != 0:
                raise ValueError("Image resolution must be a multiple of 8.")

            if not is_same_res(curr_resolution, resolution):
                if is_same_res(curr_resolution, video.shape[-3:-1]):
                    video_resize = video
                else:
                    video_resize = jax.image.resize(
                        video, video.shape[0:2] + resolution + (3,), method="bilinear"
                    )

                curr_resolution = resolution
                resnet_out = hk.BatchApply(self.resnet)(
                    video_resize, is_training=is_training
                )
                latent = resnet_out["resnet_unit_3"]
                hires = resnet_out["resnet_unit_1"]
                latent = latent / jnp.sqrt(
                    jnp.maximum(
                        jnp.sum(jnp.square(latent), axis=-1, keepdims=True),
                        1e-12,
                    )
                )
                hires = hires / jnp.sqrt(
                    jnp.maximum(
                        jnp.sum(jnp.square(hires), axis=-1, keepdims=True),
                        1e-12,
                    )
                )

            feature_grid.append(latent)
            hires_feats.append(hires)
            resize_im_shape.append(video_resize.shape[2:4])

        return FeatureGrids(
            tuple(feature_grid), tuple(hires_feats), tuple(resize_im_shape)
        )

    def get_query_features(
        self,
        video: chex.Array,
        is_training: bool,
        query_points: chex.Array,
        feature_grids: Optional[FeatureGrids] = None,
        refinement_resolutions: Optional[List[Tuple[int, int]]] = None,
    ) -> QueryFeatures:
        """Computes query features, which can be used for estimate_trajectories.

        Args:
          video: A 5-D tensor representing a batch of sequences of images.
          is_training: Whether we are training.
          query_points: The query points for which we compute tracks.
          feature_grids: If passed, we'll use these feature grids rather than
            computing new ones.
          refinement_resolutions: A list of (height, width) tuples.  Refinement will
            be repeated at each specified resolution, in order to achieve high
            accuracy on resolutions higher than what TAPIR was trained on. If None,
            reasonable refinement resolutions will be inferred from the input video
            size.

        Returns:
          A QueryFeatures object which contains the required features for every
            required resolution.
        """

        if feature_grids is None:
            feature_grids = self.get_feature_grids(
                video,
                is_training=is_training,
                refinement_resolutions=refinement_resolutions,
            )

        feature_grid = feature_grids.lowres
        hires_feats = feature_grids.hires
        resize_im_shape = feature_grids.resolutions

        shape = video.shape

        # shape is [batch_size, time, height, width, channels]; conversion needs
        # [time, width, height]
        curr_resolution = (-1, -1)
        query_feats = []
        hires_query_feats = []
        for i, resolution in enumerate(resize_im_shape):
            if is_same_res(curr_resolution, resolution):
                query_feats.append(query_feats[-1])
                hires_query_feats.append(hires_query_feats[-1])
                continue
            position_in_grid = transforms.convert_grid_coordinates(
                query_points,
                shape[1:4],
                feature_grid[i].shape[1:4],
                coordinate_format="tyx",
            )
            position_in_grid_hires = transforms.convert_grid_coordinates(
                query_points,
                shape[1:4],
                hires_feats[i].shape[1:4],
                coordinate_format="tyx",
            )
            if self.parallelize_query_extraction:
                # Extracting query features involves gathering features across frames;
                # a naive implementation will cause the model to do an all gather of
                # the full video feature tensor, which consumes lots of memory.
                # Therefore, we instead perform the x,y gather for every point on every
                # single frame, and then mask out the gathers on the incorrect frames.
                # This could be made more efficient by gathering exactly one query
                # feature per device (rather than per frame).
                #
                # interp_features is now [batch, time, num_points, features]
                interp_features = jax.vmap(
                    jax.vmap(
                        jax.vmap(
                            model_utils.interp,
                            in_axes=(2, None),
                            out_axes=-1,
                        ),
                        in_axes=(0, None),
                    )
                )(feature_grid[i], position_in_grid[..., 1:])
                # is_correct_frame is [batch, time, num_points]
                frame_id = jnp.array(
                    jnp.round(position_in_grid[:, jnp.newaxis, :, 0]), jnp.int32
                )
                is_correct_frame = jax.nn.one_hot(
                    frame_id, feature_grid[i].shape[1], axis=1
                )
                interp_features = jnp.sum(
                    interp_features * is_correct_frame[..., jnp.newaxis], axis=1
                )
                hires_interp = jax.vmap(
                    jax.vmap(
                        jax.vmap(
                            model_utils.interp,
                            in_axes=(2, None),
                            out_axes=-1,
                        ),
                        in_axes=(0, None),
                    )
                )(hires_feats[i], position_in_grid_hires[..., 1:])
                hires_interp = jnp.sum(
                    hires_interp * is_correct_frame[..., jnp.newaxis], axis=1
                )
            else:
                interp_features = jax.vmap(
                    jax.vmap(
                        model_utils.interp,
                        in_axes=(3, None),
                        out_axes=1,
                    )
                )(feature_grid[i], position_in_grid)

                hires_interp = jax.vmap(
                    jax.vmap(
                        model_utils.interp,
                        in_axes=(3, None),
                        out_axes=1,
                    )
                )(hires_feats[i], position_in_grid_hires)

            hires_query_feats.append(hires_interp)
            query_feats.append(interp_features)

        return QueryFeatures(
            tuple(query_feats), tuple(hires_query_feats), tuple(resize_im_shape)
        )

    def estimate_trajectories(
        self,
        video_size: Tuple[int, int],
        is_training: bool,
        feature_grids: FeatureGrids,
        query_features: QueryFeatures,
        query_points_in_video: Optional[chex.Array],
        query_chunk_size: Optional[int] = None,
        causal_context: Optional[Sequence[Mapping[str, chex.Array]]] = None,
        get_causal_context: bool = False,
    ) -> Mapping[str, Any]:
        """Estimates trajectories given features for a video and query features.

        Args:
          video_size: A 2-tuple containing the original [height, width] of the
            video.  Predictions will be scaled with respect to this resolution.
          is_training: Whether we are training.
          feature_grids: a FeatureGrids object computed for the given video.
          query_features: a QueryFeatures object computed for the query points.
          query_points_in_video: If provided, assume that the query points come from
            the same video as feature_grids, and therefore constrain the resulting
            trajectories to (approximately) pass through them.
          query_chunk_size: When computing cost volumes, break the queries into
            chunks of this size to save memory.
          causal_context: if running online, this will be features computed for each
            trajectory on earlier frames.  If None, no context is assumed.
          get_causal_context: if True, the output dict will also include features
            that can be passed as causal_context to future frames when running
            online.

        Returns:
          A dict of outputs, including:
            occlusion: Occlusion logits, of shape [batch, num_queries, num_frames]
              where higher indicates more likely to be occluded.
            tracks: predicted point locations, of shape
              [batch, num_queries, num_frames, 2], where each point is [x, y]
              in raster coordinates
            expected_dist: uncertainty estimate logits, of shape
              [batch, num_queries, num_frames], where higher indicates more likely
              to be far from the correct answer.
            causal_context: (if get_causal_context is True) a pytree which can
              be passed as causal_context to subsequent calls if running the model
              online.  In the current model, it is a list of dicts, one per
              PIPs refinement iteration, where for each dict the values are hidden
              units from the temporal depthwise conv layers, and the keys are names
              derived from Haiku's layer names.
        """

        def train2orig(x):
            return transforms.convert_grid_coordinates(
                x,
                self.initial_resolution[::-1],
                video_size[::-1],
                coordinate_format="xy",
            )

        occ_iters = []
        pts_iters = []
        expd_iters = []
        new_causal_context = []
        num_iters = self.num_pips_iter * (len(feature_grids.lowres) - 1)
        # This contains both middle step points and final step points.
        for _ in range(num_iters + 1):
            occ_iters.append([])
            pts_iters.append([])
            expd_iters.append([])
            new_causal_context.append([])
        del new_causal_context[-1]

        infer = functools.partial(
            self.tracks_from_cost_volume,
            im_shp=feature_grids.lowres[0].shape[0:2] + self.initial_resolution + (3,),
        )

        num_queries = query_features.lowres[0].shape[1]

        # Note: the permutation is required in order to randomize which tracks
        # get the stop_gradient.
        if causal_context is None:
            perm = jax.random.permutation(hk.next_rng_key(), num_queries)
        else:
            if is_training:
                # Need to handle permutation if we want to train with causal context.
                raise ValueError("Training with causal context is not supported.")
            perm = jnp.arange(num_queries, dtype=jnp.int32)
        inv_perm = jnp.zeros_like(perm)
        inv_perm = inv_perm.at[perm].set(jnp.arange(num_queries))

        for ch in range(0, num_queries, query_chunk_size):
            perm_chunk = perm[ch : ch + query_chunk_size]
            chunk = query_features.lowres[0][:, perm_chunk]
            if causal_context is not None:
                cc_chunk = jax.tree_map(
                    lambda x: x[:, perm_chunk], causal_context
                )  # pylint: disable=cell-var-from-loop
            if query_points_in_video is not None:
                infer_query_points = query_points_in_video[
                    :, perm[ch : ch + query_chunk_size]
                ]
                num_frames = feature_grids.lowres[0].shape[1]
                infer_query_points = transforms.convert_grid_coordinates(
                    infer_query_points,
                    (num_frames,) + video_size,
                    (num_frames,) + self.initial_resolution,
                    coordinate_format="tyx",
                )
            else:
                infer_query_points = None
            points, occlusion, expected_dist = infer(
                chunk,
                feature_grids.lowres[0],
                infer_query_points,
            )
            pts_iters[0].append(train2orig(points))
            occ_iters[0].append(occlusion)
            expd_iters[0].append(expected_dist)

            mixer_feats = None

            for i in range(num_iters):
                feature_level = i // self.num_pips_iter + 1
                queries = [
                    query_features.hires[feature_level][:, perm_chunk],
                    query_features.lowres[feature_level][:, perm_chunk],
                ]
                for _ in range(self.pyramid_level):
                    queries.append(queries[-1])
                pyramid = [
                    feature_grids.hires[feature_level],
                    feature_grids.lowres[feature_level],
                ]
                for _ in range(self.pyramid_level):
                    pyramid.append(
                        hk.avg_pool(
                            pyramid[-1], [1, 1, 2, 2, 1], [1, 1, 2, 2, 1], "VALID"
                        )
                    )

                # Note: even when the pyramids are higher resolution, the points are
                # all scaled according to the original resolution.  This is because
                # the raw points are input into the model, and need to be scaled that
                # way.
                #
                # TODO(doersch): this should constrain the output to match the query
                # points.
                cc = cc_chunk[i] if causal_context is not None else None
                refined = self.refine_pips(
                    queries,
                    None,
                    pyramid,
                    points,
                    occlusion,
                    expected_dist,
                    orig_hw=self.initial_resolution,
                    last_iter=mixer_feats,
                    mixer_iter=i,
                    resize_hw=feature_grids.resolutions[feature_level],
                    causal_context=cc,
                    get_causal_context=get_causal_context,
                )
                if ch > 0:
                    refined = jax.lax.stop_gradient(refined)
                points = refined[0]
                occlusion = refined[1]
                expected_dist = refined[2]
                mixer_feats = refined[3]
                new_causal_context[i].append(refined[4])

                pts_iters[i + 1].append(train2orig(points))
                occ_iters[i + 1].append(occlusion)
                expd_iters[i + 1].append(expected_dist)

                if (i + 1) % self.num_pips_iter == 0:
                    mixer_feats = None
                    expected_dist = expd_iters[0][-1]
                    occlusion = occ_iters[0][-1]

        occlusion = []
        points = []
        expd = []

        for i in range(len(occ_iters)):
            occlusion.append(jnp.concatenate(occ_iters[i], axis=1)[:, inv_perm])
            points.append(jnp.concatenate(pts_iters[i], axis=1)[:, inv_perm])
            expd.append(jnp.concatenate(expd_iters[i], axis=1)[:, inv_perm])

        for i in range(len(new_causal_context)):
            new_causal_context[i] = jax.tree_map(
                lambda *x: jnp.concatenate(x, axis=1)[:, inv_perm],
                *new_causal_context[i],
            )

        out = dict(
            occlusion=occlusion,
            tracks=points,
            expected_dist=expd,
        )
        if get_causal_context:
            out["causal_context"] = new_causal_context

        return out

    def __call__(
        self,
        video: chex.Array,
        is_training: bool,
        query_points: chex.Array,
        query_chunk_size: Optional[int] = None,
        get_query_feats: bool = False,
        refinement_resolutions: Optional[List[Tuple[int, int]]] = None,
    ) -> Mapping[str, chex.Array]:
        """Runs a forward pass of the model.

        Args:
          video: A 5-D tensor representing a batch of sequences of images.
          is_training: Whether we are training.
          query_points: The query points for which we compute tracks.
          query_chunk_size: When computing cost volumes, break the queries into
            chunks of this size to save memory.
          get_query_feats: Return query features for other losses like contrastive.
            Not supported in the current version.
          refinement_resolutions: A list of (height, width) tuples.  Refinement will
            be repeated at each specified resolution, in order to achieve high
            accuracy on resolutions higher than what TAPIR was trained on. If None,
            reasonable refinement resolutions will be inferred from the input video
            size.

        Returns:
          A dict of outputs, including:
            occlusion: Occlusion logits, of shape [batch, num_queries, num_frames]
              where higher indicates more likely to be occluded.
            tracks: predicted point locations, of shape
              [batch, num_queries, num_frames, 2], where each point is [x, y]
              in raster coordinates
            expected_dist: uncertainty estimate logits, of shape
              [batch, num_queries, num_frames], where higher indicates more likely
              to be far from the correct answer.
        """
        if get_query_feats:
            raise ValueError("Get query feats not supported in TAPIR.")

        feature_grids = self.get_feature_grids(
            video,
            is_training,
            refinement_resolutions,
        )
        query_features = self.get_query_features(
            video,
            is_training,
            query_points,
            feature_grids,
            refinement_resolutions,
        )

        trajectories = self.estimate_trajectories(
            video.shape[-3:-1],
            is_training,
            feature_grids,
            query_features,
            query_points,
            query_chunk_size,
        )

        # The prediction is the average of the iterative refinement output across
        # every resolution.  At training time there's only one iteration, so we'll
        # just get the final refined output.  At test time, however, there will be
        # more.  The lowest resolution is at index self.num_pips_iter;
        # self.num_pips_iter iterations after that will be the next resolution,
        # and so on.
        p = self.num_pips_iter
        out = dict(
            occlusion=jnp.mean(jnp.stack(trajectories["occlusion"][p::p]), axis=0),
            tracks=jnp.mean(jnp.stack(trajectories["tracks"][p::p]), axis=0),
            expected_dist=jnp.mean(
                jnp.stack(trajectories["expected_dist"][p::p]), axis=0
            ),
            unrefined_occlusion=trajectories["occlusion"][:-1],
            unrefined_tracks=trajectories["tracks"][:-1],
            unrefined_expected_dist=trajectories["expected_dist"][:-1],
        )

        return out
